// ===============================================================
// ---------------------- NEW SPELLS PLUGIN ----------------------
// ===============================================================

#define _CRT_SECURE_NO_WARNINGS
#include <bitset>
#include "..\..\HotA\homm3.h"

#define RELEASE

#define SPELLS_MAX 128
#define ANIMS_MAX 128

#define ORIG_ADVSPELLS_NUM 10
#define ORIG_SPELLS_NUM 70
#define ARTIFACTS_NUM 144
#define HEROES_NUM 156

#define SPELLS_NUM 94
#define SPL_DISEASE 73
#define SPL_AGE 75
#define SPL_DISPEL_HELPFUL_SPELLS 78
#define SPL_FEAR 81
#define SPL_DEATH_CLOUD_NEW 82
#define SPL_DEATH_BLOW 83
#define SPL_DRAIN_LIFE 84
#define SPL_SPRITE 85
#define SPL_MAGIC_ELEMENTAL 86
#define SPL_FIREBIRD 87
#define SPL_MOBILITY 88
#define SPL_EYE_OF_THE_MAGI 89
#define SPL_TOUGHNESS 90
#define SPL_BEHEMOTHS_CLAWS 91
#define SPL_INCINERATION 92
#define SPL_EXPLOSION 93

#define SPELL_MEMORIZED 1
#define SPELL_TEMPORARY 2

#define ANIMS_NUM 89
#define ANIM_FEAR 83
#define ANIM_DEATH_CLOUD 84
#define ANIM_DEATH_BLOW 85
#define ANIM_TOUGHNESS 86
#define ANIM_BEHEMOTHS_CLAWS 87
#define ANIM_INCINERATION 88

enum e_School_Level {SL_NONE, SL_BASIC, SL_ADVANCED, SL_EXPERT};
enum e_Town {CASTLE, RAMPART, TOWER, INFERNO, NECROPOLIS, DUNGEON, STRONGHOLD, FORTRESS, CONFLUX};

#define CeilDiv(x, y) ((x - 1) / y + 1)

typedef std::bitset<ORIG_SPELLS_NUM> _SpellBitset70_;
typedef std::bitset<SPELLS_NUM> _SpellBitset_;

Patcher* _P;
PatcherInstance* _PI;

char iniPath[MAX_PATH];
char lngPath[MAX_PATH];
char lodPath[MAX_PATH];
char sndPath[MAX_PATH];

char* gmnMsk = "*.ngm?";
char* cgmMsk = "*.ncgm";
char* tgmMsk = "*.ntgm";
char* gmnExt = ".NGM%d";
char* cgmExt = ".NCGM";
char* gmnStr = "%s.NGM%d";
char* cgmStr = "NCGM";
char* tgmStr = "NTGM";

struct Game : public _GameMgr_
{
   inline int getCurrentDay()
   {
	  return curr_day_ix + (curr_week_ix - 1) * 7 + (curr_month_ix - 1) * 28;
   }
   inline bool SpellDisabled(int SpellID)
   {
	  return PField<bool>(4)[SpellID];
   }
   inline void DisableSpell(int SpellID, bool Disabled = true)
   {
	  PField<bool>(4)[SpellID] = Disabled;
   }
   inline void OpenArea(int x, int y, int z, int player, int radius, int a1)
   {
	  CALL_7(void, __thiscall, 0x49CDD0, this, x, y, z, player, radius, a1);
   }
   void SetSpellsAvailability();
   int FillShrine(int shrineFlags);
};

struct Army : public _BattleStack_
{
   inline int GetEffectiveAttackAgainst(Army* enemy, bool isRangedAttack)
   {
	  return CALL_3(int, __thiscall, 0x442130, this, enemy, isRangedAttack);
   }
   inline void CancelIndividualSpell(int SpellID)
   {
	  CALL_2(void, __thiscall, 0x444230, this, SpellID);
   }
   inline int current_health()
   {
	  return count_current * full_hp - lost_hp;
   }
   inline int full_health()
   {
	  return count_current * full_hp;
   }
   inline int fightValue()
   {
	  return creature.fight_value * count_current;
   }
   inline int GetOrientation()
   {
	  return field_44;
   }
   inline int get_total_combat_value(int a1, int a2)
   {
	  return CALL_3(int, __thiscall, 0x442B80, this, a1, a2);
   }
   bool CanUnitAttack(Army* enemy);
   int GetEffectiveDefenseAgainst(Army* enemy, bool needFrenzyMod);
   double GetDefenseModifier(Army* defender, bool isRangedAttack);
};

struct ArmyGroup : public _Army_
{
};

struct Hero : public _Hero_
{
   inline int GetSpellSchoolLevel(int SpellID, int landModifier)
   {
	  return CALL_3(int, __thiscall, 0x4E52F0, this, SpellID, landModifier); 
   }
   inline int GetManaCost(int SpellID, const ArmyGroup* armyGroup, int landModifier)
   {
	  return CALL_4(int, __thiscall, 0x4E54B0, this, SpellID, armyGroup, landModifier);
   }
   inline void SpendMana(int mana)
   {
	  CALL_2(void, __thiscall, 0x4D9540, this, mana);
   }
   void AddSpell(int SpellID);
   void UpdateSpellsFromArtifacts();
   int getBestMagicSchoolForSpell(int SpellID);
};

struct CombatManager : public _BattleMgr_
{
   inline Army* GetActiveStack()
   {
	  return (Army*)&stack[unk_side][current_stack_ix];
   }
   inline bool hasObstaclePenalty(Army* stack, int hex, int foeHex)
   {
	  return CALL_4(bool, __thiscall, 0x4670F0, this, stack, hex, foeHex);
   }
   inline bool hasDistancePenalty(Army* stack, Army* enemy)
   {
	  return CALL_3(bool, __thiscall, 0x4671E0, this, stack, enemy);
   }
   inline void CastAreaSpell(int a1, int SpellID, int a3, int a4)
   {
	  CALL_5(void, __thiscall, 0x5A4C80, this, a1, SpellID, a3, a4);
   }
   inline void PlayAnimation(int animId, Army* stack, int time = 100, bool needHitAnim = false)
   {
	  CALL_5(void, __thiscall, 0x4963C0, this, animId, stack, time, needHitAnim);
   }
   inline bool MoatPresent()
   {
	  return Field<bool>(0x53A8);
   }
   inline bool InMoat(int hex, int wallIndex)
   {
	  return CALL_3(bool, __thiscall, 0x4699A0, this, hex, wallIndex);
   }
   inline int GetAction()
   {
	  return Field<int>(0x3C);
   }
   bool CanAISummonCreature(int SpellID, int side);
};

struct AdventureManager : public _AdvMgr_
{
   inline void RedrawMiniMap(int a1, int a2, int a3, int a4, int a5)
   {
	  CALL_6(void, __thiscall, 0x4136F0, this, a1, a2, a3, a4, a5); 
   }
   inline void SimulateMouseOver(int x, int y)
   {
	  CALL_3(void, __thiscall, 0x40E2C0, this, x, y);
   }
};

struct WindowManager : public _Struct_
{
};

struct MouseManager : public _Struct_
{
   inline void SetMouseCursor(int frame, int type)
   {
	  CALL_3(void, __thiscall, 0x50CEA0, this, frame, type);
   }
   inline void TurnMouseOn(bool on = true)
   {
	  CALL_2(void, __thiscall, 0x50D7B0, this, on);
   }
};

void GetMouseCoordinates(int& x, int& y)
{
   CALL_2(void, __stdcall, 0x50D700, &x, &y);
}

// LOD
int lodType[] = {5, 1, 3};

#define pGame (*(Game**)0x699538)
#define pAdventureManager (*(AdventureManager**)0x6992B8)
#define pCombatManager (*(CombatManager**)0x699420)
#define pWindowManager (*(WindowManager**)0x6992D0)
#define pMouseManager (*(MouseManager**)0x6992B0)

struct _ComboArtInfo_
{
   int index;
   std::bitset<160> parts;
};

struct _MagicAnim_
{
   char* defName;
   char* name;
   int type;
};

struct _BookSpell_
{
   int id;
   int school_flags;
   int school_level;
};

#define o_ComboArtInfo (*(_ComboArtInfo_**)0x660B6C)
#define o_MagicAnim ((_MagicAnim_*)0x641E18)

_Spell_ Spell[SPELLS_MAX];
_MagicAnim_ SpellAnim[ANIMS_MAX];

// These are not exactly optional tables, so don't remove them
char spellIndirectTableA[] =
{
   // Spells (starting from Quicksand #10)
    0,  1,  2,  2,  3,  4,  4,  4,  4,  4,
    5,  5,  5,  5,  4,  4,  4,  4,  4,  4,
    4,  4,  4,  4,  4,  4,  4,  4,  4,  4,
    6,  4,  4,  4,  4,  4,  4,  4,  4,  4,
    4,  4,  4,  4,  4,  4,  4,  4,  4,  5,
    4,  4,  4,  7,  8,  9, 10, 10, 10, 10,
   // Special Abilities (starting from Stone Gaze #71)
    4,  4,  4,  4,  4,  4,  4,  4,  4,  4,
    4,
   // New Spells (starting from Fear #81)
    4,  5,  4,  4, 10, 10, 10,  4,  4,  4,
	4,  5,  4
};

char spellIndirectTableB[] =
{
   // Spells (starting from Quicksand #10)
    0,  1,  2,  3,  4,  5,  6,  7,  8,  9,
   10, 11, 12, 13, 14, 15, 16, 17, 17, 17,
   17, 17, 17, 17, 17, 18, 17, 19, 20, 20,
   21, 17, 17, 22, 17, 17, 17, 23, 17, 17,
   17, 17, 17, 17, 17, 17, 17,  7, 17, 24,
   17, 17, 17, 25, 26, 27, 28, 29, 30, 31,
   // Special Abilities (starting from Stone Gaze #71)
   32, 33, 34, 17, 17, 34, 37, 37, 35, 37,
   36,
   // New Spells (starting from Fear #81)
   17, 11, 17, 17, 38, 39, 40, 17, 17, 17,
   17, 11, 17
};

char spellIndirectTableC[] =
{
   // Spells (starting from Shield #27)
    0,  1,  2,  3,  4,  5,  6,  7, 32,  8,
   32, 32, 32, 32,  9, 10, 11, 12, 13, 14,
   32, 15, 16, 17, 18, 19, 20, 21, 22, 23,
   32, 24, 25, 26, 27, 28, 32, 32, 32, 32,
   32, 32, 32,
   // Special Abilities (starting from Stone Gaze #71)
   32, 29, 32, 30, 32, 31, 32, 32, 32, 32,
   32,
   // New Spells (starting from Fear #81)
   32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
   32, 32, 32
};

char spellIndirectTableD[] =
{
   // Spells (starting from Weakness #45)
    0,  1,  9,  2,  9,  9,  9,  9,  3,  4,
    9,  9,  9,  9,  9,  5,  9,  9,  9,  9,
    9,  9,  9,  9,  9,
    // Special Abilities (starting from Stone Gaze #71)
    9,  9,  6,  7,  9,  8,  9,  9,  9,  9,
    9,
    // New Spells (starting from Fear #81)
    9,  9,  9,  9,  9,  9,  9,  9,  9,  9,
	9,  9,  9
};

char spellIndirectTableE[] =
{
   // Spells (starting from Lightning Bolt #17)
    0, 16,  0, 16, 16, 16, 16,  1,  2, 16,
   16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
   16,  3,  4, 16,  5,  6, 16,  7, 16, 16,
   16, 16,  8,  8,  9,  9, 16, 16,  9, 16,
   16, 16, 10, 11, 12, 13, 16, 16, 16, 16,
   16, 16, 16,
   // Special Abilities (starting from Stone Gaze #71)
   14, 15, 16, 16, 16, 16, 16, 16, 16, 16,
   16,
   // New Spells (starting from Fear #81)
   16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
   16, 16, 16
};

char spellIndirectTableF[] =
{
   // Spells (starting from Earthquake #14)
    0,  9,  9,  9,  9,  1,  2,  2,  2,  2,
    3,  3,  3,  9,  9,  9,  9,  9,  9,  9,
    9,  4,  9,  9,  5,  5,  6,  9,  9,  9,
    9,  9,  9,  9,  9,  9,  9,  9,  9,  9,
    9,  9,  9,  9,  9,  9,  9,  9,  9,  7,
    9,  9,  8,  8,  8,  8,
   // Special Abilities (starting from Stone Gaze #71)
    9,  9,  9,  9,  9,  9,  9,  9,  9,  9,
    9,
   // New Spells (starting from Fear #81)
    9,  2,  9,  9,  8,  8,  8,  9,  9,  9,
	9,  2,  9
};

bool shrineSpells[SPELLS_MAX];

// Battle Spells
struct SlowSpell
{
   int speedMod;
}
slowSpell[2][21];

struct FearSpell
{
   int expertise;
   int speedMod;
   int retalDamageMod;
}
fearSpellParams[4], fearSpell[2][21];

struct PoisonSpell
{
   int expertise;
   float healthModFirstCast;
   float healthMod;
   float minHealth;
}
poisonSpellParams[4], poisonSpell[2][21];

struct DiseaseSpell
{
   int expertise;
   int attackPenalty;
   int defensePenalty;
   int speedMod;
}
diseaseSpellParams[4], diseaseSpell[2][21];

struct AgeSpell
{
   int healthMod;
}
ageSpell[2][21];

struct DeathBlowSpell
{
   int expertise;
}
deathBlowSpell[2][21];

struct DrainLifeSpell
{
   int expertise;
}
drainLifeSpell[2][21];

struct ToughnessSpell
{
   int healthMod;
}
toughnessSpell[2][21];

struct BehemothsClawsSpell
{
   int expertise;
}
behemothsClawsSpell[2][21];

// Adventure Spells
struct HeroAdvInfoEx
{
   int mobilityCastCount;
   int eyeOfTheMagiCastCount;
};

HeroAdvInfoEx heroAdvInfoEx[HEROES_NUM];

// Debug
void showDebugInfo(const char* format, ...)
{
   va_list args;
   va_start(args, format);

   vsprintf(o_TextBuffer, format, args);
   b_MsgBox(o_TextBuffer, MBX_OK);
   
   va_end(args);
}

void playSound(const char* fileName)
{
    CALL_1(_Sample_, __fastcall, 0x59A770, fileName);
}

char* readIniStr(const char* section, const char* key, const char* defaultStr, const char* path)
{
   const int MAX_BUFFER = 601;
   char buffer[MAX_BUFFER];
     
   GetPrivateProfileStringA(section, key, defaultStr, buffer, sizeof(buffer), path);
      
   std::string tmpStr = buffer;
   const std::string subStrA = "\\n";
   const std::string subStrB = "\n";

   int n = 0;
   while ((n = tmpStr.find(subStrA, n)) != std::string::npos)
   {
      tmpStr.replace(n, subStrA.size(), subStrB);
      n += subStrB.size();
   }
         
   char* str = (char*)malloc(tmpStr.size() + 1);
   strcpy(str, tmpStr.c_str());

   return str;
}

int creatureSpell[] =
{
   SPL_STONE, SPL_BIND, SPL_PARALYZE, SPL_DEATH_CLOUD, SPL_THUNDERBOLT,
   SPL_DISPEL_HELPFUL_SPELLS, SPL_DEATH_STARE, SPL_ACID_BREATH
};

const int newSpell[] =
{
   SPL_POISON, SPL_DISEASE, SPL_AGE, SPL_FEAR, SPL_DEATH_CLOUD_NEW,
   SPL_DEATH_BLOW, SPL_DRAIN_LIFE, SPL_SPRITE, SPL_MAGIC_ELEMENTAL, SPL_FIREBIRD,
   SPL_MOBILITY, SPL_EYE_OF_THE_MAGI, SPL_TOUGHNESS, SPL_BEHEMOTHS_CLAWS, SPL_INCINERATION,
   SPL_EXPLOSION
};

const char* oldSpellName[] =
{
   "Summon Boat", "Scuttle Boat", "Visions", "View Earth", "Disguise",
   "View Air", "Fly", "Water Walk", "Dimension Door", "Town Portal",
   "Quicksand", "Land Mine", "Force Field", "Fire Wall", "Earthquake",
   "Magic Arrow", "Ice Bolt", "Lightning Bolt", "Implosion", "Chain Lightning",
   "Frost Ring", "Fireball", "Inferno", "Meteor Shower", "Death Ripple",
   "Destroy Undead", "Armageddon", "Shield", "Air Shield", "Fire Shield",
   "Protection from Air", "Protection from Fire", "Protection from Water", "Protection from Earth", "Anti-Magic",
   "Dispel", "Magic Mirror", "Cure", "Resurrection", "Animate Dead",
   "Sacrifice", "Bless", "Curse", "Bloodlust", "Precision",
   "Weakness", "Stone Skin", "Disrupting Ray", "Prayer", "Mirth",
   "Sorrow", "Fortune", "Misfortune", "Haste", "Slow",
   "Slayer", "Frenzy", "Titan's Lightning Bolt", "Counterstrike", "Berserk",
   "Hypnotize", "Forgetfulness", "Blind", "Teleport", "Remove Obstacle",
   "Clone", "Sunmmon Fire Elemental", "Summon Earth Elemental", "Summon Water Elemental", "Summon Air Elemental"
};

const char* newSpellName[] =
{
   "Poison", "Disease", "Age", "Fear", "Death Cloud",
   "Death Blow", "Drain Life", "Summon Sprite", "Summon Magic Elemental", "Summon Firebird",
   "Mobility", "Eye of the Magi", "Toughness", "Behemoth's Claws", "Incineration",
   "Explosion"
};

int forceFixedDuration[SPELLS_NUM] = {0};

void fillSpell(const int id)
{
   int Earth, Water, Fire, Air;
   
   Spell[newSpell[id]].type = GetPrivateProfileIntA(newSpellName[id], "<Type>", 1, iniPath);
   Spell[newSpell[id]].wav_name = readIniStr(newSpellName[id], "<Sound.FileName>", "Bless.wav", iniPath);
   Spell[newSpell[id]].animation_ix = GetPrivateProfileIntA(newSpellName[id], "<Animation.ID>", 0x24, iniPath);
   Spell[newSpell[id]].flags = GetPrivateProfileIntA(newSpellName[id], "<Attributes>", 0x40845, iniPath);
   Spell[newSpell[id]].name = readIniStr(newSpellName[id], "<AbbreviatedName>", "Spell", lngPath);
   Spell[newSpell[id]].short_name = readIniStr(newSpellName[id], "<Name>", "Spell", lngPath);
   Spell[newSpell[id]].level = GetPrivateProfileIntA(newSpellName[id], "<Level>", 1, iniPath);
   Earth = GetPrivateProfileIntA(newSpellName[id], "<School.Earth>", 0, iniPath);
   Water = GetPrivateProfileIntA(newSpellName[id], "<School.Water>", 0, iniPath);
   Fire = GetPrivateProfileIntA(newSpellName[id], "<School.Fire>", 0, iniPath);
   Air = GetPrivateProfileIntA(newSpellName[id], "<School.Air>", 0, iniPath);
   Spell[newSpell[id]].school_flags = (Earth == 1) * SSF_EARTH + (Water == 1) * SSF_WATER + (Fire == 1) * SSF_FIRE + (Air == 1) * SSF_AIR;
   Spell[newSpell[id]].mana_cost[SL_NONE] = GetPrivateProfileIntA(newSpellName[id], "<Cost.None>", 5, iniPath);
   Spell[newSpell[id]].mana_cost[SL_BASIC] = GetPrivateProfileIntA(newSpellName[id], "<Cost.Basic>", 4, iniPath);
   Spell[newSpell[id]].mana_cost[SL_ADVANCED] = GetPrivateProfileIntA(newSpellName[id], "<Cost.Advanced>", 4, iniPath);
   Spell[newSpell[id]].mana_cost[SL_EXPERT] = GetPrivateProfileIntA(newSpellName[id], "<Cost.Expert>", 4, iniPath);
   Spell[newSpell[id]].eff_power = GetPrivateProfileIntA(newSpellName[id], "<Effect.Power>", 0, iniPath);
   Spell[newSpell[id]].effect[SL_NONE] = GetPrivateProfileIntA(newSpellName[id], "<Effect.None>", 0, iniPath);
   Spell[newSpell[id]].effect[SL_BASIC] = GetPrivateProfileIntA(newSpellName[id], "<Effect.Basic>", 0, iniPath);
   Spell[newSpell[id]].effect[SL_ADVANCED] = GetPrivateProfileIntA(newSpellName[id], "<Effect.Advanced>", 1, iniPath);
   Spell[newSpell[id]].effect[SL_EXPERT] = GetPrivateProfileIntA(newSpellName[id], "<Effect.Expert>", 1, iniPath);
   Spell[newSpell[id]].chance2get_var[CASTLE] = GetPrivateProfileIntA(newSpellName[id], "<ChanceToGain.Castle>", 10, iniPath);
   Spell[newSpell[id]].chance2get_var[RAMPART] = GetPrivateProfileIntA(newSpellName[id], "<ChanceToGain.Rampart>", 10, iniPath);
   Spell[newSpell[id]].chance2get_var[TOWER] = GetPrivateProfileIntA(newSpellName[id], "<ChanceToGain.Tower>", 10, iniPath);
   Spell[newSpell[id]].chance2get_var[INFERNO] = GetPrivateProfileIntA(newSpellName[id], "<ChanceToGain.Inferno>", 10, iniPath);
   Spell[newSpell[id]].chance2get_var[NECROPOLIS] = GetPrivateProfileIntA(newSpellName[id], "<ChanceToGain.Necropolis>", 10, iniPath);
   Spell[newSpell[id]].chance2get_var[DUNGEON] = GetPrivateProfileIntA(newSpellName[id], "<ChanceToGain.Dungeon>", 10, iniPath);
   Spell[newSpell[id]].chance2get_var[STRONGHOLD] = GetPrivateProfileIntA(newSpellName[id], "<ChanceToGain.Stronghold>", 10, iniPath);
   Spell[newSpell[id]].chance2get_var[FORTRESS] = GetPrivateProfileIntA(newSpellName[id], "<ChanceToGain.Fortress>", 10, iniPath);
   Spell[newSpell[id]].chance2get_var[CONFLUX] = GetPrivateProfileIntA(newSpellName[id], "<ChanceToGain.Conflux>", 10, iniPath);
   Spell[newSpell[id]].ai_value[SL_NONE] = GetPrivateProfileIntA(newSpellName[id], "<AIValue.None>", 10, iniPath);
   Spell[newSpell[id]].ai_value[SL_BASIC] = GetPrivateProfileIntA(newSpellName[id], "<AIValue.Basic>", 10, iniPath);
   Spell[newSpell[id]].ai_value[SL_ADVANCED] = GetPrivateProfileIntA(newSpellName[id], "<AIValue.Advanced>", 11, iniPath);
   Spell[newSpell[id]].ai_value[SL_EXPERT] = GetPrivateProfileIntA(newSpellName[id], "<AIValue.Expert>", 11, iniPath);
   Spell[newSpell[id]].description[SL_NONE] =
      readIniStr(newSpellName[id], "<RightClickText.Normal>", "{Spell}\n\nSpell Description.\n", lngPath);
   Spell[newSpell[id]].description[SL_BASIC] =
      readIniStr(newSpellName[id], "<RightClickText.Basic>", "{Basic Spell}\n\nBasic Spell Description.\n", lngPath);
   Spell[newSpell[id]].description[SL_ADVANCED] =
      readIniStr(newSpellName[id], "<RightClickText.Advanced>", "{Advanced Spell}\n\nAdvanced Spell Description.\n", lngPath);
   Spell[newSpell[id]].description[SL_EXPERT] =
      readIniStr(newSpellName[id], "<RightClickText.Expert>", "{Expert Spell}\n\nExpert Spell Description.\n", lngPath);
   forceFixedDuration[newSpell[id]] = GetPrivateProfileIntA(newSpellName[id], "<ForceFixedDuration>", 0, iniPath);
}

void initHeroAdvInfoEx()
{
   for (int i = 0; i < HEROES_NUM; ++i)
   {
	  heroAdvInfoEx[i].mobilityCastCount = 0;
	  heroAdvInfoEx[i].eyeOfTheMagiCastCount = 0;
   }
}

int __stdcall afterInit(LoHook* h, HookContext* c)
{
   // Spell Table
   for (int i = SPL_SUMMON_BOAT; i <= SPL_ACID_BREATH; ++i)
      Spell[i] = o_Spell[i];

   for (std::size_t i = 0; i < sizeof(newSpell) / sizeof(int); ++i)
      fillSpell(i);

   // Poison
   poisonSpellParams[SL_NONE].healthModFirstCast = 0.1f;
   poisonSpellParams[SL_NONE].healthMod = 0.1f;
   poisonSpellParams[SL_NONE].minHealth = 0.7f;
   poisonSpellParams[SL_BASIC].healthModFirstCast = 0.1f;
   poisonSpellParams[SL_BASIC].healthMod = 0.1f;
   poisonSpellParams[SL_BASIC].minHealth = 0.7f;
   poisonSpellParams[SL_ADVANCED].healthModFirstCast = 0.2f;
   poisonSpellParams[SL_ADVANCED].healthMod = 0.1f;
   poisonSpellParams[SL_ADVANCED].minHealth = 0.6f;
   poisonSpellParams[SL_EXPERT].healthModFirstCast = 0.3f;
   poisonSpellParams[SL_EXPERT].healthMod = 0.1f;
   poisonSpellParams[SL_EXPERT].minHealth = 0.5f;
   
   // Disease
   diseaseSpellParams[SL_NONE].attackPenalty = 2;
   diseaseSpellParams[SL_NONE].defensePenalty = 2;
   diseaseSpellParams[SL_NONE].speedMod = 20;
   diseaseSpellParams[SL_BASIC].attackPenalty = 2;
   diseaseSpellParams[SL_BASIC].defensePenalty = 2;
   diseaseSpellParams[SL_BASIC].speedMod = 20;
   diseaseSpellParams[SL_ADVANCED].attackPenalty = 4;
   diseaseSpellParams[SL_ADVANCED].defensePenalty = 4;
   diseaseSpellParams[SL_ADVANCED].speedMod = 40;
   diseaseSpellParams[SL_EXPERT].attackPenalty = 4;
   diseaseSpellParams[SL_EXPERT].defensePenalty = 4;
   diseaseSpellParams[SL_EXPERT].speedMod = 40;

   // Fear
   fearSpellParams[SL_NONE].retalDamageMod = 50;
   fearSpellParams[SL_BASIC].retalDamageMod = 50;
   fearSpellParams[SL_ADVANCED].retalDamageMod = 25;
   fearSpellParams[SL_EXPERT].retalDamageMod = 0;

   _PI->WriteDword(0x687FA8, (int)&Spell);

   // Magic Animation Table
   for (int i = 0; i < ANIM_FEAR; ++i)
      SpellAnim[i] = o_MagicAnim[i];

   // Fear
   SpellAnim[ANIM_FEAR].defName = "C07spE0.def";
   SpellAnim[ANIM_FEAR].name = "Fear";
   SpellAnim[ANIM_FEAR].type = 0x101;

   // Death Cloud
   SpellAnim[ANIM_DEATH_CLOUD].defName = "sp04_.def";
   SpellAnim[ANIM_DEATH_CLOUD].name = "DeathCloud";
   SpellAnim[ANIM_DEATH_CLOUD].type = 1;

   // Death Blow
   SpellAnim[ANIM_DEATH_BLOW].defName = "DeathBlow.def";
   SpellAnim[ANIM_DEATH_BLOW].name = "DeathBlow";
   SpellAnim[ANIM_DEATH_BLOW].type = 1;

   // Toughness
   SpellAnim[ANIM_TOUGHNESS].defName = "Toughness.def";
   SpellAnim[ANIM_TOUGHNESS].name = "Toughness";
   SpellAnim[ANIM_TOUGHNESS].type = 1;

   // Behemoth's Claws
   SpellAnim[ANIM_BEHEMOTHS_CLAWS].defName = "Claws.def";
   SpellAnim[ANIM_BEHEMOTHS_CLAWS].name = "Claws";
   SpellAnim[ANIM_BEHEMOTHS_CLAWS].type = 1;

   // Incineration
   SpellAnim[ANIM_INCINERATION].defName = "C13SPF0.def";
   SpellAnim[ANIM_INCINERATION].name = "Incineration";
   SpellAnim[ANIM_INCINERATION].type = 1;

   // Adventure Spells
   initHeroAdvInfoEx();

   // Animations
   int AnimAddrDefName[] = {0x43F77E, 0x43FB6A, 0x4963FB, 0x4965CF, 0x5A5036, 0x5A6B14, 0x5A7A74, 0x5A962C};
   int AnimAddrName[] = {0x4689C4, 0x49651A, 0x4966CD, 0x5A6D2D, 0x5A7B06};
   
   for (std::size_t i = 0; i < sizeof(AnimAddrDefName) / sizeof(int); ++i)
      _PI->WriteDword(AnimAddrDefName[i], (int)&SpellAnim->defName);

   for (std::size_t i = 0; i < sizeof(AnimAddrName) / sizeof(int); ++i)
      _PI->WriteDword(AnimAddrName[i], (int)&SpellAnim->name);

   _PI->WriteDword(0x43E503, (int)&SpellAnim->type);
    
   return EXEC_DEFAULT;
}

// ===============================================================
// ------------------- Availability of spells --------------------
// ---------------------------------------------------------------

// Optional
_SpellBitset_ RMGDisabledSpells;
int __stdcall RMGDisableSpells(LoHook* h, HookContext* c)
{
   for (int iSpell = 0; iSpell < SPELLS_NUM; ++iSpell)
	  RMGDisabledSpells.set(iSpell, pGame->SpellDisabled(iSpell));

   CALL_3(void, __thiscall, *(int*)(*(int*)c->ebx + 8), c->ebx, &RMGDisabledSpells, CeilDiv(ORIG_SPELLS_NUM/*SPELLS_MAX*/, 8));

   c->return_address = 0x54AF1E;
   return NO_EXEC_DEFAULT;
}

int __stdcall RMGDisableSpellsInScrollsA(LoHook* h, HookContext* c)
{
   int SpellID = (c->eax - (int)&o_Spell->school_flags) / sizeof(_Spell_);

   if (pGame->SpellDisabled(SpellID))
   {
      c->return_address = 0x5353EE;
      return NO_EXEC_DEFAULT;
   }

   return EXEC_DEFAULT;
}

int __stdcall RMGDisableSpellsInScrollsB(LoHook* h, HookContext* c)
{
   int SpellID = (c->eax - (int)&o_Spell->school_flags) / sizeof(_Spell_);

   if (pGame->SpellDisabled(SpellID))
   {
      c->return_address = 0x535423;
      return NO_EXEC_DEFAULT;
   }

   return EXEC_DEFAULT;
}

void Game::SetSpellsAvailability()
{
   char TextBuffer[64];

   // Old spells
   for (int iSpell = SPL_SUMMON_BOAT; iSpell <= SPL_AIR_ELEMENTAL; ++iSpell)
   {
	  int key = GetPrivateProfileIntA("Enabled Spells", TextBuffer, 1, iniPath);
	  sprintf(TextBuffer, "<%s>", oldSpellName[iSpell]);
	  if (key == 0)
		 DisableSpell(iSpell);
	  if (key == -1)
		 DisableSpell(iSpell, false);
   }

   // Creature spells
   for (std::size_t i = 0; i < sizeof(creatureSpell) / sizeof(int); ++i)
	  DisableSpell(creatureSpell[i]);

   // New spells
   for (std::size_t i = 0; i < sizeof(newSpell) / sizeof(int); ++i)
	  DisableSpell(newSpell[i], GetPrivateProfileIntA(newSpellName[i], "<Enabled>", 1, iniPath) != 1);
}

void Hero::AddSpell(int SpellID)
{
   spell[SpellID] = SPELL_MEMORIZED;
}

void __fastcall AddSpell(Hero* hero, int unused_edx, int SpellID)
{
   hero->AddSpell(SpellID);
}

// Spells which may appear
_SpellBitset_ mayAppearSpellsEx;
int __stdcall expandMayAppearSpells(LoHook* h, HookContext* c)
{
   int mayAppearSpellsAddr;
   int outputBitsetAddr;
   int returnAddr;
   
   switch (h->GetAddress())
   {
   case 0x5BE518:
	  mayAppearSpellsAddr = c->edi + 0xD4;
	  outputBitsetAddr = (int)&c->ecx;
	  returnAddr = 0x5BE51E;
	  break;
   case 0x5BE52D:
	  mayAppearSpellsAddr = c->edi + 0xD4;
	  outputBitsetAddr = (int)&c->eax;
	  returnAddr = 0x5BE533;
	  break;
   default:
	  mayAppearSpellsAddr = c->esi + 0xD4;
	  outputBitsetAddr = (int)&c->ebx;
	  returnAddr = 0x5BEA2A;
   }
      
   _SpellBitset70_* mayAppearSpells = (_SpellBitset70_*)mayAppearSpellsAddr;
      
   for (int iSpell = 0; iSpell < ORIG_SPELLS_NUM; ++iSpell)
	  mayAppearSpellsEx.set(iSpell, mayAppearSpells->test(iSpell)); 

   for (int iSpell = ORIG_SPELLS_NUM; iSpell < SPELLS_NUM; ++iSpell)
	  if (pGame->SpellDisabled(iSpell))
		 mayAppearSpellsEx.set(iSpell);
   
   *(int*)outputBitsetAddr = (int)&mayAppearSpellsEx;
   c->return_address = returnAddr;
   return NO_EXEC_DEFAULT;
}

// Add here *new* spells which must appear in all towns for testing purposes
int mustAppearSpellsList[] = {ID_NONE};

bool mustSpellAppearInTowns(int SpellID)
{
   bool mustAppear = false;

   for (std::size_t i = 0; i < sizeof(mustAppearSpellsList) / sizeof(int); ++i)
   {
      if (SpellID == mustAppearSpellsList[i])
      {
         mustAppear = true;
         break;
      }
   }

   return mustAppear;
}

// Spells which must appear
_SpellBitset_ mustAppearSpellsEx;
int __stdcall expandMustAppearSpells(LoHook* h, HookContext* c)
{
   _SpellBitset70_* mustAppearSpells = (_SpellBitset70_*)c->eax;
   int SpellID = c->esi;
   
   for (int iSpell = 0; iSpell < ORIG_SPELLS_NUM; ++iSpell)
	  mustAppearSpellsEx.set(iSpell, mustAppearSpells->test(iSpell)); 

   for (int iSpell = ORIG_SPELLS_NUM; iSpell < SPELLS_NUM; ++iSpell)
	  if (!pGame->SpellDisabled(iSpell) && mustSpellAppearInTowns(iSpell))
		 mustAppearSpellsEx.set(iSpell);

   c->return_address = mustAppearSpellsEx.test(SpellID) ? 0x5BEB3B : 0x5BEB1D;
   return NO_EXEC_DEFAULT;
}
// ===============================================================

int __stdcall initSpells(LoHook* h, HookContext* c)
{
   pGame->SetSpellsAvailability();  
   memcpy(&shrineSpells, &pGame->disabled_shrines, sizeof(shrineSpells) / sizeof(bool));   

   c->return_address = 0x4C2641;
   return NO_EXEC_DEFAULT;
}

int Game::FillShrine(int shrineFlags)
{
   int availSpellsNum = 0;

   for (int i = 0; i < SPELLS_NUM; ++i)
   {
      int lvl = o_Spell[i].level - 1;
      if (lvl < 5 && 1 << lvl & shrineFlags && o_Spell[i].school_flags && !shrineSpells[i])
         ++availSpellsNum;
   }

   int SpellID = 0;
   int chosenSpell = 0;

   if (availSpellsNum)
   {
      int rndSpell = Randint(0, availSpellsNum - 1);
     
      for (int i = 0; i < SPELLS_NUM; ++i)
      {
         int lvl = o_Spell[i].level - 1;
         if (lvl < 5 && 1 << lvl & shrineFlags && o_Spell[i].school_flags && !shrineSpells[i])
         {
            if (SpellID == rndSpell) break;
            ++SpellID;
         }
         ++chosenSpell;
      }
     
      shrineSpells[chosenSpell] = true;
   }
   else
   {
      SpellID = 0;
     
      for (int i = 0; i < SPELLS_NUM; ++i)
      {
         int lvl = o_Spell[i].level - 1;
         if (lvl < 5 && 1 << lvl & shrineFlags && o_Spell[i].school_flags && !disabled_shrines[i])
         {
            shrineSpells[i] = false;
            ++SpellID;
         }
      }
     
      chosenSpell = SpellID ? FillShrine(shrineFlags) : ID_NONE;
   }

   return chosenSpell;
}

int __fastcall FillShrine(Game* game, int unused_edx, int shrineFlags)
{
   return game->FillShrine(shrineFlags);
}

bool Army::CanUnitAttack(Army* enemy)
{
   if (!enemy || this == enemy)
      return false;

   if (active_spell_duration[SPL_BERSERK] || enemy->active_spell_duration[SPL_BERSERK])
      return true;

   int ownerSide = active_spell_duration[SPL_HYPNOTIZE] ? 1 - this->side : this->side;

   return ownerSide != enemy->side;
}

bool __fastcall CanUnitAttack(Army* stack, int unused_edx, Army* enemy)
{
   return stack->CanUnitAttack(enemy);
}

int Army::GetEffectiveDefenseAgainst(Army* enemy, bool needFrenzyMod)
{
   if (needFrenzyMod && active_spell_duration[SPL_FRENZY])
	  return 0;

   int effDefense = creature.defence;

   if (enemy)
   {
	  double defenseMod = 1.0;

	  if (enemy->creature_id == CID_BEHEMOTH)
		 defenseMod = 0.6;

	  if (enemy->creature_id == CID_ANCIENT_BEHEMOTH)
		 defenseMod = 0.2;

	  if (enemy->active_spell_duration[SPL_BEHEMOTHS_CLAWS] && !enemy->CanShoot())
		 defenseMod = o_Spell[SPL_BEHEMOTHS_CLAWS].effect[behemothsClawsSpell[enemy->side][enemy->index_on_side].expertise] / 100.0;

	  effDefense = (int)(effDefense * defenseMod);	  
   }

   if (pCombatManager->MoatPresent())
   {
	  int secondHexShift = GetOrientation() ? 1 : -1;
	  int second_hex_ix = creature.flags & BCF_2HEX_WIDE ? hex_ix + secondHexShift : -1;
	  
	  if (pCombatManager->InMoat(hex_ix, 0) || pCombatManager->InMoat(second_hex_ix, 0))
		 effDefense -= 3;
   }

   return effDefense;
}

int __fastcall GetEffectiveDefenseAgainst(Army* stack, int unused_edx, Army* enemy, bool needFrenzyMod)
{
   return stack->GetEffectiveDefenseAgainst(enemy, needFrenzyMod);   
}

void __stdcall beforeNewDayStart(HiHook* h, Game* game)
{
   initHeroAdvInfoEx();
   CALL_1(void, __thiscall, h->GetDefaultFunc(), game) ;
}

void __stdcall beforeNewGameStart(HiHook* h, Game* game, int a2)
{
   initHeroAdvInfoEx();
   CALL_2(void, __thiscall, h->GetDefaultFunc(), game, a2) ;
}

int __stdcall initSpellMods(LoHook* h, HookContext* c)
{
   for (int side = ATTACKER; side <= DEFENDER; ++side)
      for (int i = 0; i < 21; ++i)
      {
         pCombatManager->stack[side][i].Field<float>(0x4C8) = 1.0f;
         slowSpell[side][i].speedMod = 100;
         fearSpell[side][i].speedMod = 100;
         diseaseSpell[side][i].speedMod = 100;
         diseaseSpell[side][i].expertise = -1;
         ageSpell[side][i].healthMod = 100;
         poisonSpell[side][i].expertise = -1;
         deathBlowSpell[side][i].expertise = -1;
		 toughnessSpell[side][i].healthMod = 100;
		 behemothsClawsSpell[side][i].expertise = -1;
      }
   
   return EXEC_DEFAULT;
}

int __stdcall newRoundSpellSettings(LoHook* h, HookContext* c)
{
   Army* stack = (Army*)c->esi;

   // Fear
   if (stack->active_spell_duration[SPL_FEAR] && fearSpell[stack->side][stack->index_on_side].expertise == SL_EXPERT)
      stack->retaliations = 0;
   
   // Poison
   c->return_address = 0x447033;

   if (stack->active_spell_duration[SPL_POISON])
   {
      int schoolLevel = poisonSpell[stack->side][stack->index_on_side].expertise;
      float healthMul = max(stack->Field<float>(0x4A4) - poisonSpellParams[schoolLevel].healthMod, poisonSpellParams[schoolLevel].minHealth);
      stack->Field<float>(0x4A4) = healthMul;

      double fullHealth = (double)stack->full_hp * healthMul;

      // Poison, Age & Toughness
      if (stack->active_spell_duration[SPL_AGE])
         fullHealth *= ageSpell[stack->side][stack->index_on_side].healthMod / 100.0;

	  if (stack->active_spell_duration[SPL_TOUGHNESS])
		 fullHealth *= toughnessSpell[stack->side][stack->index_on_side].healthMod / 100.0;

      int health = (int)(fullHealth + 0.95);

      bool needPoisonAnim = stack->creature.hit_points > health;
      if (needPoisonAnim)
		 c->return_address = 0x446FCB;

      stack->creature.hit_points = health;

      if (health - 1 < stack->lost_hp)
         stack->lost_hp = health - 1;
   }
      
   return NO_EXEC_DEFAULT;
}

int __stdcall applySpell(LoHook* h, HookContext* c)
{
   int SpellID = *(int*)(c->ebp + 8);
   Hero* hero = *(Hero**)(c->ebp + 0x14);
   Army* stack = (Army*)c->esi;
   int schoolLevel = c->eax;
   int spellSpecialtyEffect = 0;
   
   switch (SpellID)
   {
   case SPL_FEAR:
      if (!(stack->creature.flags & BCF_CANT_MOVE))
      {
         fearSpell[stack->side][stack->index_on_side].expertise = schoolLevel;
         fearSpell[stack->side][stack->index_on_side].speedMod = o_Spell[SPL_FEAR].effect[schoolLevel];

         if (schoolLevel == SL_EXPERT)
         {
            stack->CancelIndividualSpell(SPL_COUNTERSTRIKE);
            stack->retaliations = 0;
         }

         int minSpeedMod = 100;

         if (stack->active_spell_duration[SPL_SLOW] && slowSpell[stack->side][stack->index_on_side].speedMod < minSpeedMod)
            minSpeedMod = slowSpell[stack->side][stack->index_on_side].speedMod;

         if (stack->active_spell_duration[SPL_DISEASE] && diseaseSpell[stack->side][stack->index_on_side].speedMod < minSpeedMod)
            minSpeedMod = diseaseSpell[stack->side][stack->index_on_side].speedMod;

         if (fearSpell[stack->side][stack->index_on_side].speedMod < minSpeedMod)
            minSpeedMod = fearSpell[stack->side][stack->index_on_side].speedMod;
               
         stack->Field<float>(0x4C8) = minSpeedMod / 100.0f;
      }
      break;

   case SPL_POISON:
      if (schoolLevel > poisonSpell[stack->side][stack->index_on_side].expertise)
      {
         float healthMod = poisonSpellParams[schoolLevel].healthModFirstCast;
         poisonSpell[stack->side][stack->index_on_side].expertise = schoolLevel;

         float healthMul = max(stack->Field<float>(0x4A4) - healthMod, poisonSpellParams[schoolLevel].minHealth);
         stack->Field<float>(0x4A4) = healthMul;

         double fullHealth = (double)stack->full_hp * healthMul;

         // Poison, Age & Toughness
         if (stack->active_spell_duration[SPL_AGE])
            fullHealth *= ageSpell[stack->side][stack->index_on_side].healthMod / 100.0;

		 if (stack->active_spell_duration[SPL_TOUGHNESS])
			fullHealth *= toughnessSpell[stack->side][stack->index_on_side].healthMod / 100.0;

         int health = (int)(fullHealth + 0.95);

         stack->creature.hit_points = health;
         
         if (health - 1 < stack->lost_hp)
            stack->lost_hp = health - 1;
      }
      break;

   case SPL_DISEASE:
      if (schoolLevel > diseaseSpell[stack->side][stack->index_on_side].expertise)
      {
         // Init
         diseaseSpell[stack->side][stack->index_on_side].attackPenalty = diseaseSpellParams[schoolLevel].attackPenalty;
         diseaseSpell[stack->side][stack->index_on_side].defensePenalty = diseaseSpellParams[schoolLevel].defensePenalty;

         // Optional, as we don't have heroes with Disease spell specialty yet
         if (hero)
         {
            spellSpecialtyEffect = CALL_4(int, __thiscall, 0x4E6260, hero, SPL_DISEASE, hero->level, 0);
            diseaseSpell[stack->side][stack->index_on_side].attackPenalty += spellSpecialtyEffect;
            diseaseSpell[stack->side][stack->index_on_side].defensePenalty += spellSpecialtyEffect;
         }

         stack->creature.attack -= diseaseSpell[stack->side][stack->index_on_side].attackPenalty;
         stack->creature.defence -= diseaseSpell[stack->side][stack->index_on_side].defensePenalty;

         // Not allowing attack and defense to drop below 0
         if (stack->creature.attack < 0)
         {
            diseaseSpell[stack->side][stack->index_on_side].attackPenalty += stack->creature.attack;
            stack->creature.attack = 0;
         }

         if (stack->creature.defence < 0)
         {
            diseaseSpell[stack->side][stack->index_on_side].defensePenalty += stack->creature.defence;
            stack->creature.defence = 0;
         }

         if (!(stack->creature.flags & BCF_CANT_MOVE))
         {
            diseaseSpell[stack->side][stack->index_on_side].speedMod =
               o_Spell[SPL_DISEASE].effect[schoolLevel] + spellSpecialtyEffect;

            int minSpeedMod = 100;

            if (stack->active_spell_duration[SPL_SLOW] && slowSpell[stack->side][stack->index_on_side].speedMod < minSpeedMod)
               minSpeedMod = slowSpell[stack->side][stack->index_on_side].speedMod;

            if (stack->active_spell_duration[SPL_FEAR] && fearSpell[stack->side][stack->index_on_side].speedMod < minSpeedMod)
               minSpeedMod = fearSpell[stack->side][stack->index_on_side].speedMod;

            if (diseaseSpell[stack->side][stack->index_on_side].speedMod < minSpeedMod)
               minSpeedMod = diseaseSpell[stack->side][stack->index_on_side].speedMod;
               
            stack->Field<float>(0x4C8) = minSpeedMod / 100.0f;
         }
      }
      break;

   case SPL_SLOW:
      if (!(stack->creature.flags & BCF_CANT_MOVE))
      {
         stack->CancelIndividualSpell(SPL_HASTE);
         slowSpell[stack->side][stack->index_on_side].speedMod = *(int*)(c->ebp + 0x10);

         int minSpeedMod = 100;

         if (stack->active_spell_duration[SPL_FEAR] && fearSpell[stack->side][stack->index_on_side].speedMod < minSpeedMod)
            minSpeedMod = fearSpell[stack->side][stack->index_on_side].speedMod;

         if (stack->active_spell_duration[SPL_DISEASE] && diseaseSpell[stack->side][stack->index_on_side].speedMod < minSpeedMod)
            minSpeedMod = diseaseSpell[stack->side][stack->index_on_side].speedMod;

         if (slowSpell[stack->side][stack->index_on_side].speedMod < minSpeedMod)
            minSpeedMod = slowSpell[stack->side][stack->index_on_side].speedMod;
               
         stack->Field<float>(0x4C8) = minSpeedMod / 100.0f;

         // Slow animation
         stack->Field<int>(0x158) = (int)(stack->Field<int>(0x68) * 1.5);
      }
      break;

   case SPL_AGE:
      {
         float healthMul = stack->Field<float>(0x4A4);
         int ageHealthMod = min(o_Spell[SPL_AGE].effect[schoolLevel], ageSpell[stack->side][stack->index_on_side].healthMod);
         ageSpell[stack->side][stack->index_on_side].healthMod = ageHealthMod;

         double fullHealth = (double)stack->full_hp * healthMul * ageHealthMod / 100.0;
         
		 if (stack->active_spell_duration[SPL_TOUGHNESS])
			fullHealth *= toughnessSpell[stack->side][stack->index_on_side].healthMod / 100.0;
		 
		 int health = (int)(fullHealth + 0.95);

         stack->creature.hit_points = health;

         if (health - 1 < stack->lost_hp)
            stack->lost_hp = health - 1;
      }
      break;

   case SPL_DEATH_BLOW:
      deathBlowSpell[stack->side][stack->index_on_side].expertise = schoolLevel;
      break;

   case SPL_DRAIN_LIFE:
      drainLifeSpell[stack->side][stack->index_on_side].expertise = schoolLevel;
      break;

   case SPL_TOUGHNESS:
      {
		 float healthMul = stack->Field<float>(0x4A4);
		 int toughnessHealthMod = o_Spell[SPL_TOUGHNESS].effect[schoolLevel];
		 toughnessSpell[stack->side][stack->index_on_side].healthMod = toughnessHealthMod;

		 double fullHealth = (double)stack->full_hp * healthMul * toughnessHealthMod / 100.0;
		 
		 if (stack->active_spell_duration[SPL_AGE])
			fullHealth *= ageSpell[stack->side][stack->index_on_side].healthMod / 100.0;

         int health = (int)(fullHealth + 0.95);

         stack->creature.hit_points = health;

         if (health - 1 < stack->lost_hp)
            stack->lost_hp = health - 1;
      }
      break;

   case SPL_BEHEMOTHS_CLAWS:
	  behemothsClawsSpell[stack->side][stack->index_on_side].expertise = schoolLevel;
	  break;

   default:
      return EXEC_DEFAULT;
   }

   // Force fixed duration
   if (forceFixedDuration[SpellID])
	  stack->active_spell_duration[SpellID] = forceFixedDuration[SpellID];
   
   c->return_address = 0x444D5C;
   return NO_EXEC_DEFAULT;
}

int __stdcall resetSpell(LoHook* h, HookContext* c)
{
   int SpellID = *(int*)(c->ebp + 8);
   Army* stack = (Army*)c->esi;

   switch (SpellID)
   {
   case SPL_FEAR:
      if (!(stack->creature.flags & BCF_CANT_MOVE))
      {
         int minSpeedMod = 100;

         if (stack->active_spell_duration[SPL_SLOW] && slowSpell[stack->side][stack->index_on_side].speedMod < minSpeedMod)
            minSpeedMod = slowSpell[stack->side][stack->index_on_side].speedMod;

         if (stack->active_spell_duration[SPL_DISEASE] && diseaseSpell[stack->side][stack->index_on_side].speedMod < minSpeedMod)
            minSpeedMod = diseaseSpell[stack->side][stack->index_on_side].speedMod;

         stack->Field<float>(0x4C8) = minSpeedMod / 100.0f;
      }

      fearSpell[stack->side][stack->index_on_side].speedMod = 100;
      break;

   case SPL_POISON:
      {
         stack->Field<float>(0x4A4) = 1.0f;
         poisonSpell[stack->side][stack->index_on_side].expertise = -1;
                 
         double fullHealth = (double)stack->full_hp;

         // Poison, Age & Toughness
         if (stack->active_spell_duration[SPL_AGE])
            fullHealth *= ageSpell[stack->side][stack->index_on_side].healthMod / 100.0;

		 if (stack->active_spell_duration[SPL_TOUGHNESS])
			fullHealth *= toughnessSpell[stack->side][stack->index_on_side].healthMod / 100.0;

         int health = (int)(fullHealth + 0.95);

         stack->creature.hit_points = health;

         if (health - 1 < stack->lost_hp)
            stack->lost_hp = health - 1;
      }
      break;

   case SPL_DISEASE:
      stack->creature.attack += diseaseSpell[stack->side][stack->index_on_side].attackPenalty;
      stack->creature.defence += diseaseSpell[stack->side][stack->index_on_side].defensePenalty;

      if (!(stack->creature.flags & BCF_CANT_MOVE))
      {
         int minSpeedMod = 100;

         if (stack->active_spell_duration[SPL_SLOW] && slowSpell[stack->side][stack->index_on_side].speedMod < minSpeedMod)
            minSpeedMod = slowSpell[stack->side][stack->index_on_side].speedMod;

         if (stack->active_spell_duration[SPL_FEAR] && fearSpell[stack->side][stack->index_on_side].speedMod < minSpeedMod)
            minSpeedMod = fearSpell[stack->side][stack->index_on_side].speedMod;

         stack->Field<float>(0x4C8) = minSpeedMod / 100.0f;
      }
      
      diseaseSpell[stack->side][stack->index_on_side].speedMod = 100;
      diseaseSpell[stack->side][stack->index_on_side].expertise = -1;
      break;
   
   case SPL_SLOW:
      if (!(stack->creature.flags & BCF_CANT_MOVE))
      {
         int minSpeedMod = 100;

         if (stack->active_spell_duration[SPL_FEAR] && fearSpell[stack->side][stack->index_on_side].speedMod < minSpeedMod)
            minSpeedMod = fearSpell[stack->side][stack->index_on_side].speedMod;

         if (stack->active_spell_duration[SPL_DISEASE] && diseaseSpell[stack->side][stack->index_on_side].speedMod < minSpeedMod)
            minSpeedMod = diseaseSpell[stack->side][stack->index_on_side].speedMod;

         stack->Field<float>(0x4C8) = minSpeedMod / 100.0f;
      }

      // Slow animation
      stack->Field<int>(0x158) = stack->Field<int>(0x68); 

      slowSpell[stack->side][stack->index_on_side].speedMod = 100;
      break;

   case SPL_AGE:
      {
         float healthMul = stack->Field<float>(0x4A4);
         ageSpell[stack->side][stack->index_on_side].healthMod = 100;

         double fullHealth = (double)stack->full_hp * healthMul;
		 
		 if (stack->active_spell_duration[SPL_TOUGHNESS])
			fullHealth *= toughnessSpell[stack->side][stack->index_on_side].healthMod / 100.0;

         int health = (int)(fullHealth + 0.95);

         stack->creature.hit_points = health;

         if (health - 1 < stack->lost_hp)
            stack->lost_hp = health - 1;
      }
      break;

   case SPL_DEATH_BLOW:
      deathBlowSpell[stack->side][stack->index_on_side].expertise = -1;
      break;

   case SPL_DRAIN_LIFE:
      drainLifeSpell[stack->side][stack->index_on_side].expertise = -1;
      break;

   case SPL_TOUGHNESS:
      {
         float healthMul = stack->Field<float>(0x4A4);
         toughnessSpell[stack->side][stack->index_on_side].healthMod = 100;

         double fullHealth = (double)stack->full_hp * healthMul;
		 
		 if (stack->active_spell_duration[SPL_AGE])
			fullHealth *= ageSpell[stack->side][stack->index_on_side].healthMod / 100.0;

         int health = (int)(fullHealth + 0.95);

         stack->creature.hit_points = health;

         if (health - 1 < stack->lost_hp)
            stack->lost_hp = health - 1;
      }
      break;

   case SPL_BEHEMOTHS_CLAWS:
	  behemothsClawsSpell[stack->side][stack->index_on_side].expertise = -1;
	  break;
   
   default:
      return EXEC_DEFAULT;
   }

   c->return_address = 0x4444E1;
   return NO_EXEC_DEFAULT;
}

int __stdcall creatureCast(LoHook* h, HookContext* c)
{
   int SpellID = c->eax;
   int hex_ix = *(int*)(c->esi - 0xB4);

   switch (SpellID)
   {
   case SPL_POISON:
   case SPL_DISEASE:
	  pCombatManager->CastSpell(SpellID, hex_ix, 1, -1, SL_NONE, 3); 
      break;
   case SPL_AGE:
	  pCombatManager->CastSpell(SpellID, hex_ix, 1, -1, SL_EXPERT, 3); 
      break;
   default:
      return EXEC_DEFAULT;
   }
   
   c->return_address = 0x468CE7;
   return NO_EXEC_DEFAULT;
}


int __stdcall zombieDiseaseOnMagicPlains(LoHook* h, HookContext* c)
{
   int SpellID = *(int*)(c->ebp + 8);
   
   if (pCombatManager->Field<int>(0x53C0) == 1 && SpellID == SPL_DISEASE)
   {
      c->esi = 1;
      *(int*)(c->ebp - 0x44) = 1;
   }
      
   return EXEC_DEFAULT;
}

int __stdcall cureNewSpells(LoHook* h, HookContext* c)
{
   Army* stack = (Army*)c->esi;

   CALL_2(void, __thiscall, 0x444230, stack, SPL_FEAR);
   
   return EXEC_DEFAULT;
}

bool isDrainLifeImmune(int CreatureID)
{
   switch (CreatureID)
   {
   case CID_STONE_GARGOYLE:
   case CID_OBSIDIAN_GARGOYLE:
   case CID_STONE_GOLEM:
   case CID_IRON_GOLEM:
   case CID_GOLD_GOLEM:
   case CID_DIAMOND_GOLEM:
   case CID_GIANT:
   case CID_TITAN:
   case CID_VAMPIRE_LORD:
   case CID_AIR_ELEMENTAL:
   case CID_STORM_ELEMENTAL:
   case CID_EARTH_ELEMENTAL:
   case CID_MAGMA_ELEMENTAL:
   case CID_FIRE_ELEMENTAL:
   case CID_ENERGY_ELEMENTAL:
   case CID_WATER_ELEMENTAL:
   case CID_ICE_ELEMENTAL:
   case CID_PSYCHIC_ELEMENTAL:
   case CID_MAGIC_ELEMENTAL:
      return true;
   default:
      return false;
   }
}

bool isSupportiveUnitNotBallista(int CreatureID)
{
   return CreatureID == CID_CATAPULT || CreatureID == CID_FIRST_AID_TENT || CreatureID == CID_AMMO_CART;
}

int __stdcall getStackMagicVulnerability(LoHook* h, HookContext* c)
{
   _CreatureInfo_* creature = (_CreatureInfo_*)c->edi;
   int CreatureID = *(int*)(c->ebp - 8);
   int SpellID = *(int*)(c->ebp - 0x10);

   bool immune = false; 

   switch (SpellID)
   {
   // Let's allow to cast Disease on battle machines
   case SPL_DISEASE:
      if (!(creature->flags & BCF_ALIVE) && !(creature->flags & BCF_CANT_MOVE))
         immune = true;
      break;
   case SPL_AGE:
      if (!(creature->flags & BCF_ALIVE))
         immune = true;
      break;
   case SPL_FEAR:
      if (!(creature->flags & BCF_ALIVE))
         immune = true;
      break;
   case SPL_DEATH_CLOUD_NEW:
	  if (creature->flags & BCF_UNDEAD)
		 immune = true;
	  break;
   case SPL_DEATH_BLOW:
      break;
   case SPL_DRAIN_LIFE:
      if (isDrainLifeImmune(CreatureID))
         immune = true;
      break;
   case SPL_BEHEMOTHS_CLAWS:
	  if (CreatureID == CID_ANCIENT_BEHEMOTH)
		 immune = true;
	  break;
   default:
      return EXEC_DEFAULT;
   }

   c->return_address = immune ? 0x44A3E8 : 0x44A416;
   return NO_EXEC_DEFAULT;
}

int Hero::getBestMagicSchoolForSpell(int SpellID)
{
   int magicSchoolSecSkill = HSS_EARTH_MAGIC;
   int maxExpertise = second_skill[HSS_EARTH_MAGIC];
   
   if (o_Spell[SpellID].flags & SSF_AIR && second_skill[HSS_AIR_MAGIC] > maxExpertise)
   {
	  maxExpertise = second_skill[HSS_AIR_MAGIC];
	  magicSchoolSecSkill = HSS_AIR_MAGIC;
   }

   if (o_Spell[SpellID].flags & SSF_FIRE && second_skill[HSS_FIRE_MAGIC] > maxExpertise)
   {
	  maxExpertise = second_skill[HSS_FIRE_MAGIC];
	  magicSchoolSecSkill = HSS_FIRE_MAGIC;
   }

   if (o_Spell[SpellID].flags & SSF_WATER && second_skill[HSS_WATER_MAGIC] > maxExpertise)
   {
	  maxExpertise = second_skill[HSS_WATER_MAGIC];
	  magicSchoolSecSkill = HSS_WATER_MAGIC;
   }

   return magicSchoolSecSkill;
}

// ===============================================================
// --------------------- AI Spell Weighting ----------------------
// ---------------------------------------------------------------

// class type_AI_spellcaster
struct type_AI_spellcaster : _Struct_
{
   _ptr_ func;
   Hero* hero;
   Hero* enemyHero;
   // ...
};

// struct type_enchant_data
struct type_enchant_data : _Struct_
{
   int SpellID;
   int a2; // School Level, but not always
   int a3; // Spell Power
   int a4; // Spell Power
   int a5;
};

void AI_debug_spellcast(type_AI_spellcaster* caster, type_enchant_data data, int SpellID, int value)
{
   showDebugInfo("{%s}\n\nValue = %d\n\nF20 = %d\nF24 = %d\nF28 = %d\n\nSpellID = %d\na2 = %d\na3 = %d\na4 = %d\na5 = %d",
	  o_Spell[SpellID].name, value, caster->Field<int>(0x20), caster->Field<int>(0x24), caster->Field<int>(0x28),
	  data.SpellID, data.a2, data.a3, data.a4, data.a5);
}

int __fastcall get_poison_value(type_AI_spellcaster* caster, int unused_edx, Army* stack, type_enchant_data data)
{
   if (caster->Field<char>(0x1C))
	  return 0;

   int value = 0;
   if (stack && caster->hero->power > 3)
   {
      bool canShoot = CALL_2(bool, __thiscall, 0x442610, stack, 0);
      value = CALL_6(int, __thiscall, 0x442CF0, stack, caster->Field<int>(0x20), caster->Field<int>(0x24), canShoot,
         (stack->creature.hit_points - stack->lost_hp) / 2, caster->Field<int>(0x28));

	  value *= max(stack->get_total_combat_value(caster->Field<int>(0x20), caster->Field<int>(0x24)) / 2000, 1);

	  // Adjustings
      if (caster->hero)
      {
		 const int SpellID = SPL_POISON;
		 double expertiseCoef[] =
		 {
			1.0 - o_Spell[SpellID].effect[SL_NONE] / 100.0,
			1.0 - o_Spell[SpellID].effect[SL_BASIC] / 100.0,
			1.0 - o_Spell[SpellID].effect[SL_ADVANCED] / 100.0,
			1.0 - o_Spell[SpellID].effect[SL_EXPERT] / 100.0
		 };
         
		 int magicSchool = caster->hero->getBestMagicSchoolForSpell(SpellID);
		 value += (int)(value * caster->hero->power * 0.5 * expertiseCoef[caster->hero->second_skill[magicSchool]]);
#ifdef DEBUG
		 AI_debug_spellcast(caster, data, SpellID, value);
#endif
      }
   }

   return value;
}

int __fastcall get_disease_value(type_AI_spellcaster* caster, int unused_edx, Army* stack, type_enchant_data data)
{
   if (caster->Field<char>(0x1C))
	  return 0;
      
   int value = 0;
   if (stack && caster->hero->power > 1)
   {
      value = stack->get_total_combat_value(caster->Field<int>(0x20), caster->Field<int>(0x24)) / 2;

      // Adjustments
      if (caster->hero)
      {
         const int SpellID = SPL_DISEASE;
		 double expertiseCoef[] =
		 {
			1.0 - o_Spell[SPL_DISEASE].effect[SL_NONE] / 100.0,
			1.0 - o_Spell[SPL_DISEASE].effect[SL_BASIC] / 100.0,
			1.0 - o_Spell[SPL_DISEASE].effect[SL_ADVANCED] / 100.0,
			(1.0 - o_Spell[SPL_DISEASE].effect[SL_EXPERT] / 100.0) * 3
		 };

         int magicSchool = caster->hero->getBestMagicSchoolForSpell(SPL_DISEASE);
		 value += (int)(value * caster->hero->power * 0.5 * expertiseCoef[caster->hero->second_skill[magicSchool]]);
#ifdef DEBUG
		 AI_debug_spellcast(caster, data, SpellID, value);
#endif
      }
   }

   return value;
}

int __fastcall get_age_value(type_AI_spellcaster* caster, int unused_edx, Army* stack, type_enchant_data data)
{
   if (caster->Field<char>(0x1C))
	  return 0;
      
   int value = 0;
   if (stack && caster->hero->power > 1)
   {
      value = stack->get_total_combat_value(caster->Field<int>(0x20), caster->Field<int>(0x24)) / 2;

      // Adjustments
      if (caster->hero)
      {
         const int SpellID = SPL_AGE;
		 double expertiseCoef[] =
		 {
			1.0 - o_Spell[SpellID].effect[SL_NONE] / 100.0,
			1.0 - o_Spell[SpellID].effect[SL_BASIC] / 100.0,
			1.0 - o_Spell[SpellID].effect[SL_ADVANCED] / 100.0,
			1.0 - o_Spell[SpellID].effect[SL_EXPERT] / 100.0
		 };
         
		 int magicSchool = caster->hero->getBestMagicSchoolForSpell(SpellID);
         value += (int)(value * caster->hero->power * 0.5  * expertiseCoef[caster->hero->second_skill[magicSchool]]);
#ifdef DEBUG
		 AI_debug_spellcast(caster, data, SpellID, value);
#endif
      }
   }

   return value;
}

int __fastcall get_fear_value(type_AI_spellcaster* caster, int unused_edx, Army* stack, type_enchant_data data)
{
   if (caster->Field<char>(0x1C))
	  return 0;

   int value = 0;
   if (stack && caster->hero->power > 1)
   {
      value = stack->get_total_combat_value(caster->Field<int>(0x20), caster->Field<int>(0x24));

      // Adjustments
      if (caster->hero)
      {
         const int SpellID = SPL_FEAR;
		 double expertiseCoef[] =
		 {
			1.0 - o_Spell[SpellID].effect[SL_NONE] / 100.0,
			1.0 - o_Spell[SpellID].effect[SL_BASIC] / 100.0,
			1.0 - o_Spell[SpellID].effect[SL_ADVANCED] / 100.0,
			1.0 - o_Spell[SpellID].effect[SL_EXPERT] / 100.0
		 };

		 int magicSchool = caster->hero->getBestMagicSchoolForSpell(SpellID);
         value += (int)(value * caster->hero->power * 0.5 * expertiseCoef[caster->hero->second_skill[magicSchool]]);
#ifdef DEBUG
		 AI_debug_spellcast(caster, data, SpellID, value);
#endif
      }
   }

   return value;
}

int __fastcall get_death_blow_value(type_AI_spellcaster* caster, int unused_edx, Army* stack, type_enchant_data data)
{
   if (caster->Field<char>(0x1C))
	  return 0;

   int value = 0;
   if (stack && caster->hero->power > 1)
   {
	  if (!(stack->creature.flags & BCF_CAN_SHOOT))
	  {
		 value = stack->get_total_combat_value(caster->Field<int>(0x20), caster->Field<int>(0x24)) / 4;

		 // Adjustments
		 if (caster->hero)
		 {
			const int SpellID = SPL_DEATH_BLOW;
			double expertiseCoef[] =
			{
			   o_Spell[SpellID].effect[SL_NONE] / 100.0,
			   o_Spell[SpellID].effect[SL_BASIC] / 100.0,
			   o_Spell[SpellID].effect[SL_ADVANCED] / 100.0,
			   o_Spell[SpellID].effect[SL_EXPERT] / 100.0
			};

			int magicSchool = caster->hero->getBestMagicSchoolForSpell(SpellID);
			value = (int)(value * caster->hero->power * 0.5 * expertiseCoef[caster->hero->second_skill[magicSchool]]);
#ifdef DEBUG
			AI_debug_spellcast(caster, data, SpellID, value);
#endif
		 }
	  }
   }

   return value;
}

int __fastcall get_drain_life_value(type_AI_spellcaster* caster, int unused_edx, Army* stack, type_enchant_data data)
{
   if (caster->Field<char>(0x1C))
	  return 0;

   int value = 0;
   if (stack && caster->hero->power > 1)
   {
	  if (!(stack->creature.flags & BCF_CAN_SHOOT))
	  {
		 value = stack->get_total_combat_value(caster->Field<int>(0x20), caster->Field<int>(0x24)) / 4;

		 // Adjustments
		 if (caster->hero)
		 {
			const int SpellID = SPL_DRAIN_LIFE;
			double expertiseCoef[] =
			{
			   o_Spell[SpellID].effect[SL_NONE] / 100.0,
			   o_Spell[SpellID].effect[SL_BASIC] / 100.0,
			   o_Spell[SpellID].effect[SL_ADVANCED] / 100.0,
			   o_Spell[SpellID].effect[SL_EXPERT] / 100.0
			};

			int magicSchool = caster->hero->getBestMagicSchoolForSpell(SpellID);
			value = (int)(value * caster->hero->power * 0.5 * expertiseCoef[caster->hero->second_skill[magicSchool]]);
#ifdef DEBUG
			AI_debug_spellcast(caster, data, SpellID, value);
#endif
		 }
	  }
   }

   return value;
}

int __fastcall get_toughness_value(type_AI_spellcaster* caster, int unused_edx, Army* stack, type_enchant_data data)
{
   if (caster->Field<char>(0x1C))
	  return 0;

   int value = 0;
   if (stack && caster->hero->power > 1)
   {
      value = stack->get_total_combat_value(caster->Field<int>(0x20), caster->Field<int>(0x24)) / 3;

      // Adjustments
      if (caster->hero)
      {
         const int SpellID = SPL_TOUGHNESS;
		 double expertiseCoef[] =
		 {
			o_Spell[SpellID].effect[SL_NONE] / 100.0 - 1.0,
			o_Spell[SpellID].effect[SL_BASIC] / 100.0 - 1.0,
			o_Spell[SpellID].effect[SL_ADVANCED] / 100.0 - 1.0,
			o_Spell[SpellID].effect[SL_EXPERT] / 100.0 - 1.0
		 };

		 int magicSchool = caster->hero->getBestMagicSchoolForSpell(SpellID);
         value = (int)(value * caster->hero->power * 0.5 * expertiseCoef[caster->hero->second_skill[magicSchool]]);
#ifdef DEBUG
		 AI_debug_spellcast(caster, data, SpellID, value);
#endif
      }
   }

   return value;
}

int __fastcall get_behemoths_claws_value(type_AI_spellcaster* caster, int unused_edx, Army* stack, type_enchant_data data)
{
   if (caster->Field<char>(0x1C))
	  return 0;

   int value = 0;
   if (stack && caster->hero->power > 1)
   {
	  if (!(stack->creature.flags & BCF_CAN_SHOOT))
	  {
		 value = stack->get_total_combat_value(caster->Field<int>(0x20), caster->Field<int>(0x24)) / 4;

		 // Adjustments
		 if (caster->hero)
		 {
			const int SpellID = SPL_BEHEMOTHS_CLAWS;
			double expertiseCoef[] =
			{
			   1.0 - o_Spell[SpellID].effect[SL_NONE] / 100.0,
			   1.0 - o_Spell[SpellID].effect[SL_BASIC] / 100.0,
			   1.0 - o_Spell[SpellID].effect[SL_ADVANCED] / 100.0,
			   1.0 - o_Spell[SpellID].effect[SL_EXPERT] / 100.0
			};
            
			int magicSchool = caster->hero->getBestMagicSchoolForSpell(SpellID);
			value = (int)(value * caster->hero->power * 0.5 * expertiseCoef[caster->hero->second_skill[magicSchool]]);
#ifdef DEBUG
			AI_debug_spellcast(caster, data, SpellID, value);
#endif
		 }
	  }
   }

   return value;
}

int __stdcall get_enchantment_function(HiHook* h, type_AI_spellcaster* caster, int SpellID)
{
   switch (SpellID)
   {
   case SPL_POISON:
      return (int)get_poison_value;
      break;
   case SPL_DISEASE:
      return (int)get_disease_value;
      break;
   case SPL_AGE:
      return (int)get_age_value;
      break;
   case SPL_FEAR:
      return (int)get_fear_value;
      break;
   case SPL_DEATH_BLOW:
      return (int)get_death_blow_value;
      break;
   case SPL_DRAIN_LIFE:
      return (int)get_drain_life_value;
      break;
   case SPL_TOUGHNESS:
	  return (int)get_toughness_value;
      break;
   case SPL_BEHEMOTHS_CLAWS:
	  return (int)get_behemoths_claws_value;
      break;
   default:
      return CALL_2(int, __thiscall, h->GetDefaultFunc(), caster, SpellID);
   }
}

bool CombatManager::CanAISummonCreature(int SpellID, int side)
{
   if (stacks_count[side] >= 20)
	  return false;

   int CreatureID = PField<int>(0x132A8)[side];
   if (CreatureID == ID_NONE)
      return true;

   bool result;

   switch (SpellID)
   {
   case SPL_FIRE_ELEMENTAL:
      result = CreatureID == CID_FIRE_ELEMENTAL;
      break;
   case SPL_EARTH_ELEMENTAL:
      result = CreatureID == CID_EARTH_ELEMENTAL;
      break;
   case SPL_WATER_ELEMENTAL:
      result = CreatureID == CID_WATER_ELEMENTAL;
      break;
   case SPL_AIR_ELEMENTAL:
      result = CreatureID == CID_AIR_ELEMENTAL;
      break;
   case SPL_SPRITE:
      result = CreatureID == CID_SPRITE;
      break;
   case SPL_MAGIC_ELEMENTAL:
      result = CreatureID == CID_MAGIC_ELEMENTAL;
      break;
   case SPL_FIREBIRD:
      result = CreatureID == CID_FIREBIRD;
      break;
   default:
      result = CreatureID == ID_NONE;
   }

   return result;
}

bool __fastcall CanAISummonCreature(CombatManager* combatMgr, int unused_edx, int SpellID, int side)
{
   return combatMgr->CanAISummonCreature(SpellID, side);   
}

int __fastcall getCreatureTypeToSummon(int SpellID)
{
   int CreatureID;

   switch (SpellID)
   {
   case SPL_FIRE_ELEMENTAL:
      CreatureID = CID_FIRE_ELEMENTAL;
      break;
   case SPL_EARTH_ELEMENTAL:
      CreatureID = CID_EARTH_ELEMENTAL;
      break;
   case SPL_WATER_ELEMENTAL:
      CreatureID = CID_WATER_ELEMENTAL;
      break;
   case SPL_AIR_ELEMENTAL:
      CreatureID = CID_AIR_ELEMENTAL;
      break;
   case SPL_SPRITE:
      CreatureID = CID_SPRITE;
      break;
   case SPL_MAGIC_ELEMENTAL:
      CreatureID = CID_MAGIC_ELEMENTAL;
      break;
   case SPL_FIREBIRD:
      CreatureID = CID_FIREBIRD;
      break;
   default:
      CreatureID = ID_NONE;
   }

   return CreatureID;
}

double getSummoningEffect(int SpellID, int schoolLevel)
{
   double effect = 0.0;
   double elem_AI_Value_Avg = (o_pCreatureInfo[CID_AIR_ELEMENTAL].AI_value + o_pCreatureInfo[CID_EARTH_ELEMENTAL].AI_value +
      o_pCreatureInfo[CID_FIRE_ELEMENTAL].AI_value + o_pCreatureInfo[CID_WATER_ELEMENTAL].AI_value) / 4.0;

   switch (SpellID)
   {
   case SPL_FIRE_ELEMENTAL:
   case SPL_EARTH_ELEMENTAL:
   case SPL_WATER_ELEMENTAL:
   case SPL_AIR_ELEMENTAL:
      effect = 1;
      break;
   case SPL_SPRITE:
      effect = 0.8 * elem_AI_Value_Avg / o_pCreatureInfo[CID_SPRITE].AI_value;
      break;
   case SPL_MAGIC_ELEMENTAL:
      effect = 1.2 * elem_AI_Value_Avg / o_pCreatureInfo[CID_MAGIC_ELEMENTAL].AI_value;
      break;
   case SPL_FIREBIRD:
      effect = 1.2 * elem_AI_Value_Avg / o_pCreatureInfo[CID_FIREBIRD].AI_value;
      break;
   }

   effect *= o_Spell[SpellID].effect[schoolLevel];

   return effect;
}

int __stdcall battleAISpellWeighting(LoHook* h, HookContext* c)
{
   int SpellID = c->edi;
   int schoolLevel = *(int*)(c->ebx + 4);
   int spellPower = *(int*)(c->ebx + 8);
   int side = *(int*)(c->esi + 0xC);

   switch (SpellID)
   {
   case SPL_FIRE_ELEMENTAL:
   case SPL_EARTH_ELEMENTAL:
   case SPL_WATER_ELEMENTAL:
   case SPL_AIR_ELEMENTAL:
   case SPL_SPRITE:
   case SPL_MAGIC_ELEMENTAL:
   case SPL_FIREBIRD:
	  if (!*(bool*)(c->esi + 0x1C) && pCombatManager->CanAISummonCreature(SpellID, side))
      {
         *(bool*)(c->ebx + 0x20) = true;
         
         int creaturesNum = (int)(getSummoningEffect(SpellID, schoolLevel) * spellPower);
         if (creaturesNum < 1) creaturesNum = 1;

         int k = *(bool*)(c->esi + 0x28) ? 1000 : creaturesNum;
         *(int*)(c->ebx + 0x1C) = k * o_pCreatureInfo[getCreatureTypeToSummon(SpellID)].AI_value;
      }
      break;
   default:
      return EXEC_DEFAULT;
   }
   
   c->return_address = 0x43B9BF;
   return NO_EXEC_DEFAULT;
}
// ===============================================================

// ===============================================================
// ---- Summon Sprite, Summon Magic Elemental, Summon Firebird ---
// ---------------------------------------------------------------
int __stdcall SummonCreatures(LoHook* h, HookContext* c)
{
   int SpellID = *(int*)(c->ebp + 8);
   int CreatureID;

   switch (SpellID)
   {
   case SPL_SPRITE:
      CreatureID = CID_SPRITE;
      break;
   case SPL_MAGIC_ELEMENTAL:
      CreatureID = CID_MAGIC_ELEMENTAL;
      break;
   case SPL_FIREBIRD:
      CreatureID = CID_FIREBIRD;
      break;
   default:
      return EXEC_DEFAULT;
   }

   CALL_5(void, __thiscall, 0x5A7390, pCombatManager, SpellID, CreatureID, *(int*)(c->ebp + 0x1C), c->esi);

   c->return_address = 0x5A2368;
   return NO_EXEC_DEFAULT;
}

int __stdcall skipNonElementals(LoHook* h, HookContext* c)
{
   int CreatureID = *(int*)(c->ebp + 0xC);
   
   switch (CreatureID)
   {
   case CID_SPRITE:
   case CID_FIREBIRD:
      c->return_address = 0x5A74DD;
      return NO_EXEC_DEFAULT;
   default:
      return EXEC_DEFAULT;
   }
}

int spriteSpellParams[] = {1000, 1000, 1500, 2000};
int magicElementalSpellParams[] = {100, 100, 150, 200};
int firebirdSpellParams[] = {50, 50, 75, 100};

int __stdcall setSummonedCreaturesNumber(LoHook* h, HookContext* c)
{
   int CreatureID = *(int*)(c->ebp + 0xC);
   int spellPower = *(int*)(c->ebp + 0x10);
   int schoolLevel = *(int*)(c->ebp + 0x14);

   switch (CreatureID)
   {
   case CID_SPRITE:
      c->esi = spriteSpellParams[schoolLevel];
      break;
   case CID_MAGIC_ELEMENTAL:
      c->esi = magicElementalSpellParams[schoolLevel];
      break;
   case CID_FIREBIRD:
      c->esi = firebirdSpellParams[schoolLevel];
      break;
   default:
      return EXEC_DEFAULT;
   }

   c->esi = c->esi * spellPower / 100;
   if (c->esi < 1) c->esi = 1;
   
   c->ecx = c->ebx;
   c->return_address = 0x5A7526;
   return NO_EXEC_DEFAULT;
}

int __stdcall checkSummonedCreaturesType(LoHook* h, HookContext* c)
{
   c->return_address = 0x59F8B7;
   int SpellID = c->ebx;
     
   switch (SpellID)
   {
   case SPL_SPRITE:
   case SPL_FIREBIRD:
      c->return_address = 0x59F8C0;
      break;
   case SPL_FIRE_ELEMENTAL:
      c->eax = CID_FIRE_ELEMENTAL;
      break;
   case SPL_EARTH_ELEMENTAL:
      c->eax = CID_EARTH_ELEMENTAL;
      break;
   case SPL_WATER_ELEMENTAL:
      c->eax = CID_WATER_ELEMENTAL;
      break;
   case SPL_AIR_ELEMENTAL:
      c->eax = CID_AIR_ELEMENTAL;
      break;
   case SPL_MAGIC_ELEMENTAL:
      c->eax = CID_MAGIC_ELEMENTAL;
      break;
   default:
      c->eax = ID_NONE;
   }
      
   return NO_EXEC_DEFAULT;
}
// ===============================================================

// ===============================================================
// ----------------------------- Fear ----------------------------
// ---------------------------------------------------------------
int __stdcall skipMeleeAttackUnderFear(LoHook* h, HookContext* c)
{
   Army* stack = pCombatManager->GetActiveStack();
   
   if (stack && stack->active_spell_duration[SPL_FEAR])
   {
      c->return_address = 0x4220B3;
      return NO_EXEC_DEFAULT;
   }

   return EXEC_DEFAULT;
}

int __stdcall skipShootingUnderFear(LoHook* h, HookContext* c)
{
   Army* stack = pCombatManager->GetActiveStack();
   
   if (stack && stack->active_spell_duration[SPL_FEAR])
   {
      c->return_address = 0x41F263;
      return NO_EXEC_DEFAULT;
   }

   return EXEC_DEFAULT;
}

double Army::GetDefenseModifier(Army* defender, bool isRangedAttack)
{
   int attack = GetEffectiveAttackAgainst(defender, isRangedAttack);
   int defense = defender->GetEffectiveDefenseAgainst(this, true);

   double mod = defense > attack ? min(1.0 - (defense - attack) * 0.025, 0.3) : 1.0;
   
   if (creature_id == CID_PSYCHIC_ELEMENTAL && (defender->creature.flags >> 10) & 1)
      mod *= 0.5;

   if (creature_id == CID_MAGIC_ELEMENTAL &&
	  (defender->creature_id == CID_MAGIC_ELEMENTAL || defender->creature_id == CID_BLACK_DRAGON))
      mod *= 0.5;

   if (isRangedAttack)
   {
      int hex = hex_ix; 
      if (creature.flags & BCF_2HEX_WIDE)
         hex += GetOrientation() ? 1 : -1;

      if (pCombatManager->hasObstaclePenalty(this, hex, defender->hex_ix))
         mod *= 0.5;

      if (pCombatManager->hasDistancePenalty(this, defender))
         mod *= 0.5;
   }

   if ((creature.flags >> 2) & 1 && !isRangedAttack && !((creature.flags >> 12) & 1))
      mod *= 0.5;

   // Effects
   bool blinded = Field<bool>(0x4C0);
   bool paralyzed = Field<bool>(0x4C1);
   int blindRetalMod = (int)(Field<float>(0x49C) * 100);
   int paralyzeRetalMod = o_Spell[SPL_PARALYZE].effect[SL_ADVANCED];

   int effectMod = 100;

   if (blinded && blindRetalMod < effectMod)
      effectMod = blindRetalMod;

   if (paralyzed && paralyzeRetalMod < effectMod)
      effectMod = paralyzeRetalMod;
   
   if (active_spell_duration[SPL_FEAR])
   {
      int schoolLevel = fearSpell[side][index_on_side].expertise;
      if (fearSpellParams[schoolLevel].retalDamageMod < effectMod)
         effectMod = fearSpellParams[schoolLevel].retalDamageMod;
   }

   mod = mod * effectMod / 100;
   return mod;
}

double __fastcall GetDefenseModifier(Army* attacker, int unused_edx, Army* defender, bool isRangedAttack)
{
   return attacker->GetDefenseModifier(defender, isRangedAttack);   
}

int __stdcall setCursorForFearSpell(HiHook* h, CombatManager* combatMgr, int hex)
{
   int cursorType = CALL_2(int, __thiscall, h->GetDefaultFunc(), combatMgr, hex);

   Army* stack = combatMgr->GetActiveStack();
      
   if (stack && stack->active_spell_duration[SPL_FEAR] && (cursorType == 3 || cursorType == 15 || cursorType == 7))
      return 0;

   return cursorType;
}
// ===============================================================

// Death Cloud, Incineration
int __stdcall castBattleSpell(LoHook* h, HookContext* c)
{
   int SpellID = c->edx;
   Army* stack = (Army*)c->ebx;
   
   switch (SpellID)
   {
   case SPL_DEATH_CLOUD_NEW:
   case SPL_INCINERATION:
	  pCombatManager->CastAreaSpell(*(int*)(c->ebp + 0xC), SpellID, c->esi, *(int*)(c->ebp + 0x1C));
	  break;
   case SPL_EXPLOSION:
	  pCombatManager->PlayAnimation(o_Spell[SPL_EXPLOSION].animation_ix, stack, 10); 
	  c->return_address = 0x5A0E2A;
	  return NO_EXEC_DEFAULT;
   default:
	  return EXEC_DEFAULT;
   }

   c->return_address = 0x5A2368;
   return NO_EXEC_DEFAULT;
}

// Death Blow
int __stdcall DeathBlow(LoHook* h, HookContext* c)
{
   Army* stack = (Army*)c->esi;
   int schoolLevel = deathBlowSpell[stack->side][stack->index_on_side].expertise;
   
   if (stack->active_spell_duration[SPL_DEATH_BLOW] && pCombatManager->GetAction() == 6 &&
	  Randint(1, 100) <= o_Spell[SPL_DEATH_BLOW].effect[schoolLevel])
   {
	  bool virtualDamage = *(bool*)&c->ebx;
      c->return_address = virtualDamage ? 0x44388E : 0x4436E7;
      return NO_EXEC_DEFAULT;
   }
   
   return EXEC_DEFAULT;
}

// Drain Life
int __stdcall DrainLife(LoHook* h, HookContext* c)
{
   Army* stack = (Army*)c->esi;
   
   if (stack->active_spell_duration[SPL_DRAIN_LIFE])
   {
      c->return_address = 0x440921;
      return NO_EXEC_DEFAULT;
   }

   return EXEC_DEFAULT;
}

int __stdcall calcDrainLifeHeal(LoHook* h, HookContext* c)
{
   Army* stack = (Army*)c->esi;

   if (stack->active_spell_duration[SPL_DRAIN_LIFE])
      c->edx = (int)(c->edx * o_Spell[SPL_DRAIN_LIFE].effect[drainLifeSpell[stack->side][stack->index_on_side].expertise] / 100.0);

   return EXEC_DEFAULT;
}

// ===============================================================
// --------------------- Adventure Spells ------------------------
// ---------------------------------------------------------------
bool eyeOfTheMagiCast = false;

int __stdcall eyeOfTheMagi(LoHook* h, HookContext* c)
{
   if (eyeOfTheMagiCast)
   {
	  c->return_address = 0x4915C0;
	  return NO_EXEC_DEFAULT;
   }

   return EXEC_DEFAULT;
}

void __stdcall setMouseCursor(HiHook* h, MouseManager* mouseManager, int frame, int type)
{
   if (eyeOfTheMagiCast && frame == 41 && type == 1)
   {
	  frame = 0;
	  type = 3;
   }

   CALL_3(void, __thiscall, h->GetDefaultFunc(), mouseManager, frame, type);
}

int mobilitySpellParams[] = {1, 2, 3, 4};
int eyeOfTheMagiSpellParams[] = {1, 2, 3, 4};

int __stdcall castAdventureSpell(LoHook* h, HookContext* c)
{
   int SpellID = *(int*)(c->ebp + 8);
   Hero* hero = (Hero*)c->esi;

   int landModifier = hero->GetLandModifierUnder();
   int schoolLevel = hero->GetSpellSchoolLevel(SpellID, landModifier);

   switch (SpellID)
   {
   case SPL_MOBILITY:
	  if (heroAdvInfoEx[hero->id].mobilityCastCount < mobilitySpellParams[schoolLevel])
	  {
		 ++heroAdvInfoEx[hero->id].mobilityCastCount;
		 hero->SpendMana(hero->GetManaCost(SPL_MOBILITY, NULL, landModifier));
		 hero->movement_points += o_Spell[SPL_MOBILITY].effect[schoolLevel];
		 hero->movement_points_max += o_Spell[SPL_MOBILITY].effect[schoolLevel];
		 CALL_4(void, __thiscall, 0x418D30, o_AdvMgr, 0, 0, 1);
		 pAdventureManager->FullUpdate(true);
		 playSound(o_Spell[SPL_MOBILITY].wav_name);
	  }
	  else
	  {
		 sprintf(o_TextBuffer, o_GENRLTXT_TXT->GetString(339), hero->name);
		 b_MsgBox(o_TextBuffer, MBX_OK);
	  }
	  break;

   case SPL_EYE_OF_THE_MAGI:
	  if (heroAdvInfoEx[hero->id].eyeOfTheMagiCastCount < eyeOfTheMagiSpellParams[schoolLevel])
	  {
		 _Dlg_* dlg = new _Dlg_();
		 CALL_1(void, __thiscall, 0x41D326 + 5 + *(int*)0x41D327, dlg);
		 eyeOfTheMagiCast = true;
		 CALL_2(void, __thiscall, 0x5FFA20, dlg, 0);
		 eyeOfTheMagiCast = false;
		 CALL_1(void, __thiscall, 0x41D346 + 5 + *(int*)0x41D347, dlg);

		 int packedMapItem;
		 CALL_2(void, __thiscall, 0x407A70, o_AdvMgr, &packedMapItem);

		 if (CALL_1(bool, __thiscall, 0x4B1090, &packedMapItem) && pWindowManager->Field<int>(0x38))
		 {
			int x = b_unpack_x(packedMapItem);
			int y = b_unpack_y(packedMapItem);
			int z = b_unpack_z(packedMapItem);
			pGame->OpenArea(x, y, z, o_ActivePlayer->id, 2, 0);
			
			_Struct_* p = (_Struct_*)pAdventureManager;
			p->Field<int>(0xEC) = -1;
			GetMouseCoordinates(x, y);
			pAdventureManager->SimulateMouseOver(x, y);
		 }
		 else
		 {
			sprintf(o_TextBuffer, o_GENRLTXT_TXT->GetString(732));
			b_MsgBox(o_TextBuffer, MBX_OK);
			break;
		 }

		 ++heroAdvInfoEx[hero->id].eyeOfTheMagiCastCount;
		 hero->SpendMana(hero->GetManaCost(SPL_EYE_OF_THE_MAGI, NULL, landModifier));
		 pAdventureManager->FullUpdate(true);
		 playSound(o_Spell[SPL_EYE_OF_THE_MAGI].wav_name);
	  }
	  else
	  {
		 sprintf(o_TextBuffer, o_GENRLTXT_TXT->GetString(339), hero->name);
		 b_MsgBox(o_TextBuffer, MBX_OK);
	  }
	  break;
   
   default:
	  return EXEC_DEFAULT;
   }
   
   c->return_address = 0x41C67B;
   return NO_EXEC_DEFAULT;
}

int __stdcall loadHeroAdvInfoEx(LoHook* h, HookContext* c)
{
   Hero* hero = (Hero*)c->edi;

   if (hero)
   {
	  CALL_3(void, __thiscall, *(int*)(*(int*)c->esi + 4), c->esi, c->ebp + 8, 4);
	  heroAdvInfoEx[hero->id].mobilityCastCount = *(int*)(c->ebp + 8);
	  CALL_3(void, __thiscall, *(int*)(*(int*)c->esi + 4), c->esi, c->ebp + 8, 4);
	  heroAdvInfoEx[hero->id].eyeOfTheMagiCastCount = *(int*)(c->ebp + 8);
   }

   return EXEC_DEFAULT;
}

int __stdcall saveHeroAdvInfoEx(LoHook* h, HookContext* c)
{
   Hero* hero = (Hero*)c->edi;

   if (hero)
   {
	  *(int*)(c->ebp + 8) = heroAdvInfoEx[hero->id].mobilityCastCount;
	  CALL_3(void, __thiscall, *(int*)(*(int*)c->esi + 8), c->esi, c->ebp + 8, 4);
	  *(int*)(c->ebp + 8) = heroAdvInfoEx[hero->id].eyeOfTheMagiCastCount;
	  CALL_3(void, __thiscall, *(int*)(*(int*)c->esi + 8), c->esi, c->ebp + 8, 4);
   }

   return EXEC_DEFAULT;
}
// ===============================================================

void Hero::UpdateSpellsFromArtifacts()
{
   for (int iSpell = 0; iSpell < SPELLS_MAX; ++iSpell)
      if (spell[iSpell] > SPELL_MEMORIZED)
		 spell[iSpell] = 0;

   for (int iSlot = 0; iSlot < 19; ++iSlot)
   {
      if (doll_art[iSlot].id != ID_NONE)
      {
         if (doll_art[iSlot].id == AID_SPELL_SCROLL)
         {
            if (doll_art[iSlot].mod < SPELLS_NUM && !spell[doll_art[iSlot].mod])
               spell[doll_art[iSlot].mod] = SPELL_TEMPORARY;
         }
         else if (o_ArtInfo[doll_art[iSlot].id].new_spell)
         {
            _SpellBitset_ spells;
            CALL_2(int, __fastcall, 0x4D95C0, &spells, doll_art[iSlot].id);

            for (std::size_t iSpell = 0; iSpell < spells.size(); ++iSpell)
            {
               if (spells[iSpell] && !spell[iSpell])
                  spell[iSpell] = SPELL_TEMPORARY;
            }

            int comboArtIndex = o_ArtInfo[doll_art[iSlot].id].supercomposite;
            if (comboArtIndex != -1)
            {
               for (int iArt = 0; iArt < ARTIFACTS_NUM; ++iArt)
               {
                  if (o_ComboArtInfo[comboArtIndex].parts[iArt] && o_ArtInfo[iArt].new_spell)
                  {
                     _SpellBitset_ spells;
                     CALL_2(int, __fastcall, 0x4D95C0, &spells, iArt);

                     for (std::size_t iSpell = 0; iSpell < spells.size(); ++iSpell)
                     {
                        if (spells[iSpell] && !spell[iSpell])
                           spell[iSpell] = SPELL_TEMPORARY;
                     }
                  }
               }
            }
         }
      }
   }
}

void __fastcall UpdateSpellsFromArtifacts(Hero* hero)
{
   hero->UpdateSpellsFromArtifacts();   
}

int __stdcall loadLod(LoHook* h, HookContext* c)
{
   int lodNumber = **(int**)(c->ebp - 0x18);
   if (lodNumber == 5)
   {
	  c->eax = (int)lodPath;
	  c->return_address = 0x55A592;
	  return NO_EXEC_DEFAULT;
   }

   return EXEC_DEFAULT;
}

int __stdcall loadSnd(LoHook* h, HookContext* c)
{
   c->eax = (int)sndPath;

   return EXEC_DEFAULT;
}

BOOL APIENTRY DllMain(HMODULE hModule, DWORD ul_reason_for_call, LPVOID lpReserved)
{
   static bool plugin_On = false;
   if ( DLL_PROCESS_ATTACH == ul_reason_for_call )
   {
      if ( !plugin_On )
      {
         plugin_On = true;
         _P = GetPatcher();
         _PI = _P->CreateInstance("HD.Plugin.H3.NewSpells");

         char curDir[MAX_PATH];
		 GetCurrentDirectoryA(sizeof(curDir), curDir);
		 sprintf(iniPath, "%s\\NewSpells.ini", curDir);
		 
		 char lngName[32];
		 GetPrivateProfileStringA("Language", "<Name>", "English", lngName, sizeof(lngName), iniPath);

		 sprintf(lngPath, "%s\\Language\\%s.ini", curDir, lngName);
		 sprintf(lodPath, "%s\\Data\\NewSpells.lod", curDir);
		 sprintf(sndPath, "%s\\Data\\NewSpells.snd", curDir);

         // Moving and Expanding Spell Table
         _PI->WriteLoHook     (0x4EE1C1, afterInit);
               
         // nwcthereisnospoon
         _PI->WriteByte       (0x402902, SPELLS_NUM);

         // "(Already learned)" Text
         _PI->WriteDword      (0x40D97C, 0x3EA);
         
         // AI
         _PI->WriteDword      (0x41FBDF, 0x3EA);
         _PI->WriteByte       (0x41FC91, SPELLS_NUM);
         _PI->WriteDword      (0x425C72, 0x3EA);
         _PI->WriteDword      (0x425E98, SPELLS_NUM * sizeof(_Spell_));
         _PI->WriteDword      (0x427039, 0x3EA);
         _PI->WriteDword      (0x427044, 0x3EA);
         _PI->WriteByte       (0x427085, SPELLS_NUM);
         _PI->WriteDword      (0x4329C1, 0x3EA);
         _PI->WriteDword      (0x432A43, SPELLS_NUM * sizeof(_Spell_));
         _PI->WriteDword      (0x432CDE, 0x3EA);
         _PI->WriteDword      (0x432D1E, SPELLS_NUM * sizeof(_Spell_));
         
         // ===============================================================
         // ----------------------- Adventure AI --------------------------
         // ---------------------------------------------------------------
         _PI->WriteDword      (0x430AE5, 0x3EA + SPL_DIMENSION_DOOR);
         _PI->WriteDword      (0x4E57CC, 0x3EA + SPL_SUMMON_BOAT);
         _PI->WriteDword      (0x56B346, 0x3EA + SPL_TOWN_PORTAL);
         _PI->WriteDword      (0x56B7EA, 0x3EA + SPL_DIMENSION_DOOR);
         _PI->WriteDword      (0x56B93F, 0x3EA + SPL_FLY);
         _PI->WriteDword      (0x56B99A, 0x3EA + SPL_WATER_WALK);
         // ===============================================================
         
         // ===============================================================
         // ------------------------ Battle AI ----------------------------
         // ---------------------------------------------------------------
         _PI->WriteDword      (0x4397E6, SPELLS_NUM * sizeof(_Spell_));
         _PI->WriteDword      (0x43C6F2, SPELLS_NUM * sizeof(_Spell_));
         // Master Genie AI Spell Weighting
         _PI->WriteDword      (0x43C21B, SPELLS_NUM * sizeof(_Spell_));
         // AI Quick Battle
         _PI->WriteDword      (0x433026, 0x3EA);
         _PI->WriteDword      (0x433719, 0x3EA);
         _PI->WriteDword      (0x43940C, 0x3EA);
         _PI->WriteDword      (0x43C561, 0x3EA);
         // AI Spell Scrolls?
         _PI->WriteDword      (0x52A97A, 0x3EA);
         _PI->WriteDword      (0x52A9B6, SPELLS_NUM * sizeof(_Spell_));
         // ===============================================================
         
         // Can cast?
         _PI->WriteDword      (0x447551, SPELLS_NUM * sizeof(_Spell_));
         
         _PI->WriteDword      (0x447C7D, SPELLS_NUM * sizeof(_Spell_));
         _PI->WriteDword      (0x447CC8, SPELLS_NUM * sizeof(_Spell_));
         
         // Cheats in battle
         _PI->WriteByte       (0x471C57, SPELLS_NUM);
         
         // Clear Hero.spell[140] (optional)
         _PI->WriteDword      (0x48647A, 0x23);
         _PI->WriteJmp        (0x486485, 0x486498);
         
         _PI->WriteByte       (0x4864B0, SPELLS_NUM);
                  
         // Scholar Secondary Skill
         _PI->WriteByte       (0x4A2743, SPELLS_NUM);

		 // Load Game
         _PI->WriteByte       (0x48A34B, SPELLS_NUM);

         _PI->WriteByte       (0x4E67AC, SPELLS_NUM);
         
         // Cheat Menu?
         _PI->WriteByte       (0x4F508B, SPELLS_NUM);
         _PI->WriteDword      (0x4F50CE, SPELLS_NUM * sizeof(_Spell_));
         _PI->WriteDword      (0x4F5114, SPELLS_NUM * sizeof(_Spell_));
                           
         // Init spells
		 _PI->WriteLoHook     (0x4C2625, initSpells);
         
		 // Shrine spells
         _PI->WriteDword      (0x4C92C5, SPELLS_NUM * sizeof(_Spell_));
         _PI->WriteDword      (0x4C9347, SPELLS_NUM * sizeof(_Spell_));
         _PI->WriteDword      (0x4C93C0, SPELLS_NUM * sizeof(_Spell_));
		 _PI->WriteHiHook     (0x4C9260, SPLICE_, DIRECT_, THISCALL_, FillShrine);

		 // Pyramids
         _PI->WriteByte       (0x4C170D, SPELLS_NUM);
                 
         _PI->WriteByte       (0x4CEC4F, SPELLS_NUM);
         
         // Clear Hero.spell[140] (optional)
         _PI->WriteDword      (0x4D8F2E, 0x23);
         _PI->WriteJmp        (0x4D8F36, 0x4D8F49);
         
         // Tome of Air Magic
         _PI->WriteDword      (0x4D962D, SPELLS_NUM * sizeof(_Spell_));
         // Tome of Fire Magic
         _PI->WriteDword      (0x4D9681, SPELLS_NUM * sizeof(_Spell_));
         // Tome of Water Magic
         _PI->WriteDword      (0x4D96D6, SPELLS_NUM * sizeof(_Spell_));
         // Tome of Earth Magic
         _PI->WriteDword      (0x4D972E, SPELLS_NUM * sizeof(_Spell_));
         // Spellbinder's Hat
         _PI->WriteDword      (0x4D9771, SPELLS_NUM * sizeof(_Spell_));
         
         // Scholars
         _PI->WriteByte       (0x5012B7, SPELLS_NUM);
         
         _PI->WriteDword      (0x527ACB, 0x3EA);
         _PI->WriteDword      (0x527B08, SPELLS_NUM * sizeof(_Spell_));
		 
		 // RMG
		 _PI->WriteDword      (0x534C4B, SPELLS_NUM * sizeof(_Spell_));
         _PI->WriteLoHook     (0x54AE0F, RMGDisableSpells);
         
         // RMG Spell Scrolls
         _PI->WriteDword      (0x5353D5, SPELLS_NUM);
         _PI->WriteByte       (0x53542B, SPELLS_NUM);
         _PI->WriteLoHook     (0x5353E8, RMGDisableSpellsInScrollsA);
         _PI->WriteLoHook     (0x535417, RMGDisableSpellsInScrollsB);
         
         // Spell Book
         _PI->WriteDword      (0x59CCDD, SPELLS_NUM * sizeof(_BookSpell_));
		 _PI->WriteDword      (0x59CD36, SPELLS_NUM * sizeof(_BookSpell_));
		 _PI->WriteDword      (0x59CD5E, 0x3EA);
         _PI->WriteDword      (0x59CDBF, SPELLS_NUM * sizeof(_Spell_));

		 // Add Spell
		 _PI->WriteHiHook     (0x4D95A0, SPLICE_, DIRECT_, THISCALL_, AddSpell);
		 
         // Cast Spell
         _PI->WriteDword      (0x5A1AD6, SPELLS_NUM * sizeof(_Spell_));

		 // ===============================================================
         // -------------------------- Mage Guild -------------------------
         // ---------------------------------------------------------------
		 _PI->WriteByte       (0x5BEA2C, SPELLS_NUM);
         _PI->WriteByte       (0x5BEA70, SPELLS_NUM);
         _PI->WriteByte       (0x5BEAAD, SPELLS_NUM);
         _PI->WriteDword      (0x5BEAFE, SPELLS_NUM * sizeof(_Spell_));
         _PI->WriteDword      (0x5BEB2C, SPELLS_NUM * sizeof(_Spell_));
         _PI->WriteByte       (0x5BEB48, SPELLS_NUM);
         _PI->WriteByte       (0x5BEB71, SPELLS_NUM);
         _PI->WriteByte       (0x5BEBB2, SPELLS_NUM);
         _PI->WriteDword      (0x5BEC05, SPELLS_NUM * sizeof(_Spell_));
         _PI->WriteByte       (0x5BEC1B, SPELLS_NUM);
		 
		 // Town setup
		 _PI->WriteByte       (0x5BEA05, 0x30); // Now we have the equivalent of std::bitset<SPELLS_MAX> on the stack
		 _PI->WriteDword      (0x5BEA12, 4); // Let's initialize it with zeroes
		 
		 int newBitsetOffset[] = {0x5BEA67, 0x5BEAB2, 0x5BEACF, 0x5BEB50, 0x5BEB68, 0x5BEBB7, 0x5BEBD6, 0x5BEC23, 0x5BEC3B};
		 for (std::size_t i = 0; i < sizeof(newBitsetOffset) / sizeof(int); ++i)
            _PI->WriteByte(newBitsetOffset[i], -0x30);

		 _PI->WriteLoHook	  (0x5BEA24, expandMayAppearSpells);
		 _PI->WriteLoHook	  (0x5BE518, expandMayAppearSpells); // Aurora Borealis
		 _PI->WriteLoHook	  (0x5BE52D, expandMayAppearSpells); // Aurora Borealis
		 _PI->WriteLoHook	  (0x5BEB15, expandMustAppearSpells);
         // ===============================================================
                 
         // Conflux Grail
         _PI->WriteDword      (0x5BE512, SPELLS_NUM * sizeof(_Spell_));
         _PI->WriteDword      (0x5BE56E, SPELLS_NUM * sizeof(_Spell_));
         _PI->WriteDword      (0x5D7464, SPELLS_NUM);
         
         // ===============================================================
         // ------------------ New Hero.spell[140] field ------------------
         // ---------------------------------------------------------------
         _PI->WriteCodePatch  (0x4D8B72, "%n", 8);
         _PI->WriteCodePatch  (0x4D8F7F, "%n", 8);
         // ===============================================================

         // ===============================================================
         // ------------- New Game.disabled_spells[140] field -------------
         // ---------------------------------------------------------------
         _PI->WriteByte       (0x4C16ED, 4); // Pyramids
         _PI->WriteByte       (0x4C254C, 4); // Titan's Lightning Bolt
         _PI->WriteByte       (0x4C25F1, 4); // Titan's Lightning Bolt
         _PI->WriteByte       (0x501297, 4); // Scholars
         _PI->WriteByte       (0x5BEA55, 4); // Mage Guild
         // ===============================================================
                         
         _PI->WriteDword      (0x52AE1C, 0x3EA);

         // ===============================================================
         // -------------------------- Battle -----------------------------
         // ---------------------------------------------------------------
         _PI->WriteByte       (0x4963E9, ANIMS_NUM);
         _PI->WriteByte       (0x4965BD, ANIMS_NUM);
         _PI->WriteByte       (0x502405, SPELLS_NUM);
         _PI->WriteByte       (0x502451, SPELLS_NUM);
         _PI->WriteByte       (0x5024CF, SPELLS_NUM);
         _PI->WriteByte       (0x502515, SPELLS_NUM);
         _PI->WriteByte       (0x502E62, SPELLS_NUM);
         _PI->WriteByte       (0x502EB7, SPELLS_NUM);
         _PI->WriteByte       (0x59EFD9, SPELLS_NUM - 1 - ORIG_ADVSPELLS_NUM);
         _PI->WriteDword      (0x59EFE4, (int)&spellIndirectTableA);
         _PI->WriteByte       (0x5A064E, SPELLS_NUM - 1 - ORIG_ADVSPELLS_NUM);
         _PI->WriteDword      (0x5A0659, (int)&spellIndirectTableB);
         _PI->WriteByte       (0x4446EA, SPELLS_NUM - 1 - SPL_SHIELD);
         _PI->WriteDword      (0x4446FD, (int)&spellIndirectTableC);
         _PI->WriteByte       (0x444262, SPELLS_NUM - 1 - SPL_WEAKNESS);
         _PI->WriteDword      (0x44427A, (int)&spellIndirectTableD);
         _PI->WriteByte       (0x44A259, SPELLS_NUM - 1 - SPL_LIGHTNING_BOLT);
         _PI->WriteDword      (0x44A264, (int)&spellIndirectTableE);
         _PI->WriteByte       (0x43B788, SPELLS_NUM - 1 - SPL_EARTHQUAKE);
         _PI->WriteDword      (0x43B793, (int)&spellIndirectTableF);
         _PI->WriteHiHook     (0x43B2E0, SPLICE_, EXTENDED_, THISCALL_, get_enchantment_function);
         _PI->WriteLoHook     (0x43B797, battleAISpellWeighting);
         // ===============================================================
                 
         // ===============================================================
         // ----- New CombatManager.active_spell_duration[162] field ------
         // ---------------------------------------------------------------
         _PI->WriteDword      (0x43787A, 162);
         _PI->WriteCodePatch  (0x437880, "%n", 19);
         _PI->WriteDword      (0x43D314, 162);
         _PI->WriteDword      (0x43E3DF, SPELLS_NUM);
         _PI->WriteByte       (0x443F63, SPELLS_NUM);
         _PI->WriteByte       (0x446F03, SPELLS_NUM);
         _PI->WriteByte       (0x5A1907, SPELLS_NUM);
         _PI->WriteByte       (0x5A1995, SPELLS_NUM);
         _PI->WriteByte       (0x5A84C8, SPELLS_NUM);
         _PI->WriteByte       (0x5A852F, SPELLS_NUM);
         _PI->WriteCodePatch  (0x4446D2, "%n", 7);
         _PI->WriteJmp        (0x444694, 0x4450C2);
         // ===============================================================
        
         // Fear, Poison, Disease, Age
         _PI->WriteLoHook     (0x444701, applySpell);
         _PI->WriteLoHook     (0x44427E, resetSpell);
         // Poison every round, no retaliations for Fear at Expert
         _PI->WriteLoHook     (0x446F11, newRoundSpellSettings);
         // Creature's cast
         _PI->WriteLoHook     (0x468CCF, creatureCast);
         // Zombie's Disease on Magic Plains
         _PI->WriteLoHook     (0x5A01E5, zombieDiseaseOnMagicPlains);
         // Allow to dispel Poison
         _PI->WriteHexPatch   (0x5A18FA, "90 90");
         _PI->WriteHexPatch   (0x5A1988, "90 90");

         // Allow to cure new spells
         _PI->WriteLoHook     (0x4462FD, cureNewSpells);

         // Death Cloud, Incineration
         _PI->WriteLoHook     (0x5A065D, castBattleSpell);

         // Death Blow
         _PI->WriteLoHook     (0x4435A9, DeathBlow);

         // Drain Life
         _PI->WriteLoHook     (0x440903, DrainLife);
         _PI->WriteLoHook     (0x44096F, calcDrainLifeHeal);

         // Summon Sprite, Summon Magic Elemental, Summon Firebird
         _PI->WriteLoHook     (0x59F889, checkSummonedCreaturesType);
         _PI->WriteLoHook     (0x5A065D, SummonCreatures);
         _PI->WriteLoHook     (0x5A74D6, skipNonElementals);
         _PI->WriteLoHook     (0x5A7516, setSummonedCreaturesNumber);
         _PI->WriteHiHook     (0x5A96B0, SPLICE_, DIRECT_, THISCALL_, CanAISummonCreature);
         _PI->WriteHiHook     (0x5A9670, SPLICE_, DIRECT_, FASTCALL_, getCreatureTypeToSummon);

		 // Adventure Spells
		 _PI->WriteLoHook     (0x41C663, castAdventureSpell);
		 _PI->WriteLoHook     (0x4915B7, eyeOfTheMagi);
		 _PI->WriteHiHook     (0x50CEA0, SPLICE_, EXTENDED_, THISCALL_, setMouseCursor);
		          
         // Stack's Magic Vulnerability
         _PI->WriteLoHook     (0x44A268, getStackMagicVulnerability);
         
         // Artifacts (incl. Spell Scrolls)
         _PI->WriteHiHook     (0x4D9840, SPLICE_, DIRECT_, THISCALL_, UpdateSpellsFromArtifacts);

         _PI->WriteHiHook     (0x4425A0, SPLICE_, DIRECT_, THISCALL_, CanUnitAttack);
		 _PI->WriteHiHook     (0x4C7CA0, SPLICE_, EXTENDED_, THISCALL_, beforeNewDayStart);
		 _PI->WriteHiHook     (0x4C2110, SPLICE_, EXTENDED_, THISCALL_, beforeNewGameStart);
		          
         // ===============================================================
         // ---------------------------- Fear -----------------------------
         // ---------------------------------------------------------------
         // Set cursor for human players
         _PI->WriteHiHook     (0x475DC0, SPLICE_, EXTENDED_, THISCALL_, setCursorForFearSpell);
         // Not allowing to attack under Fear
         _PI->WriteLoHook     (0x422088, skipMeleeAttackUnderFear);
         _PI->WriteLoHook     (0x41F25C, skipShootingUnderFear);
         // Getting defense modifier (common for Blind, Paralyze, Fear)
         _PI->WriteHiHook     (0x4422B0, SPLICE_, DIRECT_, THISCALL_, GetEffectiveDefenseAgainst);
		 _PI->WriteHiHook     (0x4438B0, SPLICE_, DIRECT_, THISCALL_, GetDefenseModifier);
         // ===============================================================

         // ===============================================================
         // ---------------------- Save & Load Game -----------------------
         // ---------------------------------------------------------------
		 _PI->WriteLoHook     (0x4D7D44, loadHeroAdvInfoEx);
		 _PI->WriteLoHook     (0x4D83B3, saveHeroAdvInfoEx);
		 
		 _PI->WriteDword	  (0x41814E, (int&)gmnExt); // .GM%d
		 _PI->WriteDword	  (0x418139, (int&)cgmExt); // .CGM
		 _PI->WriteDword	  (0x4BEC71, (int&)gmnStr); // %s.GM%d
		 _PI->WriteDword	  (0x4BEC27, (int&)cgmStr); // CGM
		 _PI->WriteDword	  (0x4BEC3B, (int&)tgmStr); // TGM
		 _PI->WriteDword	  (0x6834AC, (int&)gmnMsk); // *.gm?
		 _PI->WriteDword	  (0x6834B0, (int&)cgmMsk); // *.cgm
		 _PI->WriteDword	  (0x6834B8, (int&)tgmMsk); // *.tgm
		 // ===============================================================

		 // LOD
		 _PI->WriteDword	  (0x69E5B8, 3);
		 _PI->WriteDword	  (0x69E5B8 + 4, (int)&lodType);
		 _PI->WriteLoHook     (0x55A57D, loadLod);

		 // SND
		 _PI->WriteLoHook     (0x598979, loadSnd);
		 
		 _PI->WriteLoHook     (0x4ADF71, initSpellMods);
         
         int speedModJmpAddr[] = {0x441CE3, 0x441E06, 0x44857E, 0x4485BC, 0x4485FC, 0x44867A, 0x448A05};
		 for (std::size_t i = 0; i < sizeof(speedModJmpAddr) / sizeof(int); ++i)
            _PI->WriteHexPatch(speedModJmpAddr[i], "90 90");
      }
   }

   return TRUE;
}